#include "messiahC4Dclasses.h"
#include "messiahC4Dglobals.h"
#include "messiahC4Dmhobject.h"
#include "messiahSceneHook.h"

// old config file code, temporary
#include "ConfigData.h"

// Globals	/////////////////////////////////////////////////////////////////////////////////////////////////

extern mh::module *		messiahHost;
extern char				progName[];
extern char				plugName[];
extern ConfigData *		config;
extern bool				messiah_loaded;


// Default Ctor (::messiahMover)  ////////////////////////////////////////////////////////////////////////
messiahMover::messiahMover()
{
	// If messiah is not loaded, do so now
	// This will avoid the overhead of loading messiah until the user needs it
	if( messiah_loaded && !messiahHost->isMessiahActive() )
		if( !messiahHost->startup( progName, plugName ) )
		{
			GePrint("Error loading messiah! messiahC4D will not work properly!");
		}
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// Transform Object (::Animate)  ///////////////////////////////////////////////////////////////////////////////

LONG messiahMover::Execute(PluginTag* tag, BaseDocument* doc, BaseObject* op, BaseThread* bt, LONG priority, LONG flags)
{
	if( !messiah_loaded ) return FALSE;

	int i = 0;
	BaseObject *obj;
	BaseContainer *cont;
	BaseContainer *data;
	mC4Dmhobject *mh_obj = NULL;	// derived from mh::object

	// Get the mh::scene from the scene hook
	mh::scene* docscene = GetMessiahSceneFromDocument(*doc);
	if (!docscene) return FALSE;

	// Only interested in point objects right now
	LONG type = op->GetType();
	//if(!op->IsInstanceOf(Opoint)) return TRUE;
	
	if( !op->IsInstanceOf( Opolygon ) )
	{
		obj = op->GetCache();
		if( !obj )
		{
			ModelingCommandData cd;
			cd.doc = doc;
			cd.op = op;
			SendModelingCommand( MCOMMAND_CURRENTSTATETOOBJECT, cd );
			return TRUE;
		
		}
	}
	else
		obj = op;

	if( !( cont = tag->GetDataInstance()) ) return FALSE;
	if( !( data = cont->GetContainerInstance( ID_MESSIAH_DATA_ATTACH)) )
	// This host object has not been linked, do so now
	{
		BaseContainer newCont;
		mh_obj = new mC4Dmhobject;

		// establish link between host plugin object and host object
		// this link will be re-established each time this method is called
		// the link will then be broken upon exit
		mh_obj->d_data1 = static_cast<void *>( obj );

		// keep a link to the persistent object
		// NOTE: this object cannot be used to retrieve geom
		// added 8/12/03 - ron
		mh_obj->d_data2 = static_cast<void*>( op );

		// establish link between host object and host plugin object
		newCont.SetLong( ID_MH_OBJ_PTR, reinterpret_cast<LONG>(mh_obj) );
		cont->SetContainer( ID_MESSIAH_DATA_ATTACH, newCont );

		// establish link between host plugin object and messiah object
		String sname = op->GetName();
		char *name = new char[sname.GetCStringLen()+1];
		sname.GetCString( name, sname.GetCStringLen()+1 );
		
		docscene->connectExistingObj( mh_obj, name );
		mh_obj->setPluginType( mh::object::Motion );
		// replaced 7/12 - ron 
		// messiahHost->getScene()->connectExistingObj( mh_obj, name );

		// store update value
		mh_obj->dirty = obj->GetDirty(DIRTY_DATA);
	}
	else
	// This host object is already linked, get the mh::object from it
	{
		mh_obj = reinterpret_cast<mC4Dmhobject *>(data->GetLong( ID_MH_OBJ_PTR ));
		if(!mh_obj) return FALSE;

		if( mh_obj->loading )
		{
			if( !docscene->isLoading() ) // scene is done loading, object is connected, now rebuild the mesh
			{
				mh_obj->loading = false;
				
				// establish link between host plugin object and host object
				// this link will be re-established each time this method is called
				// the link will then be broken upon exit
				mh_obj->d_data1 = static_cast<void *>( obj );

				// keep a link to the persistent object
				// NOTE: this object cannot be used to retrieve geom
				// added 8/12/03 - ron
				mh_obj->d_data2 = static_cast<void*>( op );
				
				char const * dbg_name = mh_obj->name();
				mh_obj->rebuild(0);
				dbg_name = mh_obj->name();
				mh_obj->dirty = op->GetDirty(DIRTY_DATA);
			}
		}
		
		// Check to see if the host object's topology has changed, if so rebuild the obj
		if( op->GetDirty(DIRTY_DATA) != mh_obj->dirty )
		{
			mh_obj->d_data1 = static_cast<void *>( obj );
			mh_obj->rebuild(0);
			mh_obj->dirty = op->GetDirty(DIRTY_DATA);
		}
	}

	// get current time
	BaseTime t = doc->GetTime();
	LONG fps = doc->GetFps();
	double messiahTime = (double)t.GetFrame( fps );


	double		pos[3], rot[3];

	mh_obj->getWorldPos( pos, messiahTime );
	mh_obj->getWorldRot( rot, messiahTime );


	Vector		vpos( pos[0] / config->scaleFactor, pos[1] / config->scaleFactor, pos[2] / config->scaleFactor );


	// reconstruct local matrix

	Matrix local, h, p, b;

	double ch, cp, cb, sh, sp, sb;

	ch = cosf( rot[0] );	sh = sinf( rot[0] );
	cp = cosf( rot[1] );	sp = sinf( rot[1] );
	cb = cosf( rot[2] );	sb = sinf( rot[2] );

	h.v1.x = ch;	h.v1.z = -sh;
	h.v3.x = sh;	h.v3.z = ch;

	p.v2.y = cp;	p.v2.z = sp;
	p.v3.y = -sp;	p.v3.z = cp;

	b.v1.x = cb;	b.v1.y = sb;
	b.v2.x = -sb;	b.v2.y = cb;

	local = h * p * b;

	local.off = vpos;

	op->SetMl( local, FALSE );

//	op->SetDirty( DIRTY_MATRIX );

	return TRUE;
}

// Save Tag (::Write)  /////////////////////////////////////////////////////////////////////////////////////////

Bool messiahMover::Write(GeListNode *node, HyperFile *hf)
{
	if( !messiah_loaded ) return FALSE;

	if( !messiahHost->getScene()->isSaving() )
	{
		messiahHost->getScene()->prepareForSave();
	}

	PluginTag *tag = static_cast<PluginTag *>(node) ;
	BaseContainer *cont, *data;

	if( ( cont = tag->GetDataInstance()) )
	{
		if( ( data = cont->GetContainerInstance( ID_MESSIAH_DATA_ATTACH)) ) 
		{

			mC4Dmhobject * mh_obj = reinterpret_cast<mC4Dmhobject *>(data->GetLong( ID_MH_OBJ_PTR ));
			if(!mh_obj) return FALSE;


			hf->WriteLong( mh_obj->d_ID );

		}
	}

	return TRUE;
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Load Tag (::Read)  /////////////////////////////////////////////////////////////////////////////////////////

Bool messiahMover::Read(GeListNode *node, HyperFile *hf, LONG level )
{
	if( !messiah_loaded ) return FALSE;

	// If messiah is not loaded, do so now
	// This will avoid the overhead of loading messiah until the user needs it
	if( !messiahHost->isMessiahActive() )
		if( !messiahHost->startup( progName, plugName ) )
		{
			GePrint("Error loading messiah! messiahC4D will not work properly!");
		}

	if( !messiahHost->getScene()->isLoading() )
	{
		messiahHost->getScene()->prepareForLoad();
	}

	mC4Dmhobject *mh_obj = NULL;	// derived from mh::object

	mh_obj = new mC4Dmhobject;
	mh_obj->loading = true;

	hf->ReadLong( (LONG *)(&mh_obj->d_ID) );

	BaseTag *tag = static_cast<PluginTag *>(node);
	BaseContainer newCont, *cont;

	cont = tag->GetDataInstance();

	newCont.SetLong( ID_MH_OBJ_PTR, reinterpret_cast<LONG>(mh_obj) );
	cont->SetContainer( ID_MESSIAH_DATA_ATTACH, newCont );

	BaseObject *obj = tag->GetObject();

	// establish link between host plugin object and messiah object
	String sname = obj->GetName();
	char *name = new char[sname.GetCStringLen()+1];
	sname.GetCString( name, sname.GetCStringLen()+1 );

	messiahHost->getScene()->connectLoadingObj( mh_obj, mh_obj->d_ID );
	mh_obj->setPluginType( mh::object::Motion );

//	object_loaded( messiahHost->getScene() );

	return TRUE;
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// Register Messiah Mover Plugin  /////////////////////////////////////////////////////////////////////////////

Bool RegisterMessiahMover(void)
{
	String name;
	name = GeLoadString(IDS_MESSIAH_MOVER_TAG);

	if( !RegisterTagPlugin( ID_MESSIAH_MOVER, name, TAG_VISIBLE|TAG_EXPRESSION, messiahMover::Alloc, "", "messiah.bmp", 0) )
		return FALSE;
  
	return TRUE;
}
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

