// =============		messiahMAXclassDefs.h		==============
//
// Copyright (C) 2002 pmG Worldwide LLC. All Rights Reserved.
//
// ===============================================================
//
// Classes:
//
//		messiahMAX			- Main modifier class
//		messiahMAXClassDesc - Class descriptor 
//		messiahData			- local data
//
// ===============================================================
//
// Version History:
//
// Written By: Lyle Milton
//
// Revised By: Ron Griswold
// Date: 5/19/02
//
// ===============================================================
#ifndef MESSIAH_MAX_CLASS_DEFS
#define MESSIAH_MAX_CLASS_DEFS

#include "messiahMAX.h"
#include <vector>
#include <algorithm>
//#include <map>

using namespace std;

// Move to Params.h //////////////////////////////////////////
// ===============================================================
// Const Globals
// ===============================================================
int				const PBLOCK_REF			= 0;
Class_ID		const MESSIAHMAX_CLASS_ID	(0xf5ac1e6, 0xf04539d);
Class_ID		const MESSIAHCTRL_CLASS_ID	(0x6dd2472a, 0x55ec4a77);
Class_ID		const MESSIAHUTIL_CLASS_ID	(0x627a2075, 0x1ccc53e1);
char const *	const MMAX_INTERNAL_NAME	= "messiahMAX";

int				const MMAX_DFORMER			= FX_NODE_DISPLACEMENT; //200;
int				const MMAX_XFORMER			= FX_NODE_MOTION; //201;

int				const FILE_NODE_ID			= 100;
int				const FILE_SIZE_CHUNK		= 101;
int				const FILE_BUFF_CHUNK		= 102;
int				const FILE_NUM_NODE_REFS	= 103;
int				const FILE_NODE_REFS		= 104;

int				const MMAX_MAX_OBJ_NAME		= 256;
int				const MMAX_MAX_PTS_PER_POLY = 512;
float			const MMAX_INITIAL_SCALE	= 0.01f;
//////////////////////////////////////////////////////////////

INode* FindNodeReference( ReferenceTarget *modifier, ModContext *mod_context );
// void LogNodeData( ReferenceTarget *modifier, ReferenceTarget  *o_modifier, int *counter);
// int CollectModifiers(INode *node, int flags);


// ===============================================================
// managerElmt (Used to keep track of Mods and Controlers created for load/save)
// ===============================================================
class managerElmt	// MessiahMAX and MessiahMAXcontroller derive from this
{
public:
	managerElmt(int t) : type(t), onList(false) {}
	int	type;
	bool onList;

private:
	managerElmt();	// no default const.
};

class manager
{
public:
	std::vector<managerElmt *>	plugins;
	std::vector<managerElmt *>	loadingList;

	std::vector<ModContext *>	tmpMClist;

	typedef std::vector<managerElmt *>::iterator	plugIter;
	typedef std::vector<ModContext *>::iterator		mcIter;

	FX_NodeLog *	GetNodeLog(int messiah_obj);
	int				GetObjFromIndex(int index);

};

class enumMCs : public ModContextEnumProc
{
public:
	BOOL proc(ModContext *mc);
};

// ===============================================================
// messiahData (Local Data the Modifier and Controller)
// ===============================================================
class messiahData : public LocalModData
{	
public:
	messiahData() : messiah_object(0), loadID(0), loaded(false) {}

	int				messiah_object;
	int				loadID;
	bool			loaded;

	LocalModData *	Clone			();
	// ~messiahData() { delete messiah_object; }	// probably need to implement this
};

// ===============================================================
// messiahMAX (Modifier node)
// ===============================================================
class messiahMAX : public Modifier, public managerElmt
{
public:

	/*CST*/			messiahMAX			();
	/*DST*/			~messiahMAX			();		

	Object **		objs;
	int				numObjs;
	int *			loadIds;
	bool			loaded;

	int				getIndex	(Object * o)
	{
		for(int i = 0; i < numObjs; i++)
		{
			if(o == objs[i])
				return i;
		}
		return -1;
	}

	// Parameter block
	IParamBlock2	*pblock;	//ref 0

	static IObjParam *ip;		//Access to the interface

	//==============================================================
	// Implementation of Methods Inherited from : Animatable
	//==============================================================
	void			BeginEditParams		(IObjParam *ip, ULONG flags,Animatable *prev);
	void			EndEditParams		(IObjParam *ip, ULONG flags,Animatable *next);

	Class_ID		ClassID				()			{ return MESSIAHMAX_CLASS_ID; }		
	SClass_ID		SuperClassID		()			{ return OSM_CLASS_ID; }
	void			GetClassName		(TSTR &s)	{ s = GetString(IDS_CLASS_NAME); }

	int				NumSubs				()			{ return 1; }
	TSTR			SubAnimName			(int i)		{ return GetString(IDS_PARAMS); }				
	Animatable *	SubAnim				(int i)		{ return pblock; }

	void			DeleteThis			()			{ delete this; }		

	int				NumParamBlocks		()			{ return 1; }					
	IParamBlock2*	GetParamBlock		(int i)		{ return pblock; } 
	IParamBlock2*	GetParamBlockByID	(BlockID id){ return (pblock->ID() == id) ? pblock : NULL; } 

	//==============================================================
	// Implementation of Methods Inherited from : Modifier
	//==============================================================
	ChannelMask		ChannelsUsed		()		{ return GEOM_CHANNEL|TOPO_CHANNEL; }
	ChannelMask		ChannelsChanged		()		{ return GEOM_CHANNEL; }

	Class_ID		InputType			()		{ return defObjectClassID; } //{ return Class_ID(TRIOBJ_CLASS_ID, 0); }

	void			ModifyObject		( TimeValue t, ModContext &mc, ObjectState *os, INode *node );

	void			NotifyInputChanged	( Interval changeInt, PartID partID, RefMessage message, ModContext *mc );

	BOOL			DependOnTopology	( ModContext &mc ) {return TRUE;}

	Interval		LocalValidity		( TimeValue t );


	//==============================================================
	// Implementation of Methods Inherited from : BaseObject
	//==============================================================
	TCHAR *			GetObjectName		()		{ return GetString(IDS_CLASS_NAME); }
	BOOL			ChangeTopology		()		{ return FALSE; }		
	
	CreateMouseCallBack* 
					GetCreateMouseCallBack
										()		{ return NULL; }

	void			NotifyPreCollapse	( INode *node, IDerivedObject *derObj, int index );
	void			NotifyPostCollapse	( INode *node,Object *obj, IDerivedObject *derObj, int index );

	BOOL			HasUVW				()		{ return TRUE; }
	void			SetGenUVW			( BOOL sw );
	
	//==============================================================
	// Implementation of Methods Inherited from : ReferenceTarget
	//==============================================================
	RefTargetHandle Clone				( RemapDir &remap );
	
	//==============================================================
	// Implementation of Methods Inherited from : ReferenceMaker
	//==============================================================
	RefResult		NotifyRefChanged	(Interval changeInt, RefTargetHandle hTarget, PartID& partID,  RefMessage message);
	int				NumRefs				()		{ return 1; }
	RefTargetHandle GetReference		(int i) { return pblock; }
	void			SetReference		(int i, RefTargetHandle rtarg) { pblock=(IParamBlock2*)rtarg; }
	IOResult		Load				(ILoad *iload);
	IOResult		Save				(ISave *isave);

};

// ===============================================================
// messiahMaxClassDesc
// ===============================================================
class messiahMAXClassDesc : public ClassDesc2 
{
public:

	int 			IsPublic		()		{ return TRUE; }
	void *			Create			(BOOL loading = FALSE);
	const TCHAR *	ClassName		()		{ return GetString(IDS_CLASS_NAME); }
	SClass_ID		SuperClassID	()		{ return OSM_CLASS_ID; }
	Class_ID		ClassID			()		{ return MESSIAHMAX_CLASS_ID; }
	const TCHAR* 	Category		()		{ return GetString(IDS_CATEGORY); }
	int				BeginCreate		(Interface *i);
	int				EndCreate		(Interface *i);

	// returns fixed parsable name (scripter-visible name)
	const TCHAR*	InternalName	()		{ return _T(MMAX_INTERNAL_NAME); }	
	
	// returns owning module handle
	HINSTANCE		HInstance		()		{ return hInstance; }				

};

// ===============================================================
// messiahMAXcontroller (Transform Controller)
// ===============================================================
class messiahMAXcontroller : public StdControl, public managerElmt
{
public:

	messiahMAXcontroller();
	~messiahMAXcontroller();

	// Parameter block
	IParamBlock2	*pblock;	//ref 0

	//==============================================================
	// Implementation of Methods Inherited from : Animatable
	//==============================================================
	Class_ID		ClassID				()			{ return MESSIAHCTRL_CLASS_ID; }		
	SClass_ID		SuperClassID		()			{ return CTRL_MATRIX3_CLASS_ID; }
	void			GetClassName		(TSTR &s)	{ s = GetString(IDS_CTL_CLASS_NAME); }
	void			DeleteThis			()			{delete this;}		
	int				IsKeyable			()			{return 0;}		
	BOOL			IsAnimated			()			{return TRUE;}

	int				NumSubs				()			{ return 1; }
	TSTR			SubAnimName			(int i)		{ return GetString(IDS_PARAMS); }				
	Animatable *	SubAnim				(int i)		{ return pblock; }

	int				NumParamBlocks		()			{ return 1; }					
	IParamBlock2*	GetParamBlock		(int i)		{ return pblock; } 
	IParamBlock2*	GetParamBlockByID	(BlockID id){ return (pblock->ID() == id) ? pblock : NULL; } 

	void			BeginEditParams		(IObjParam *ip, ULONG flags,Animatable *prev);
	void			EndEditParams		(IObjParam *ip, ULONG flags,Animatable *next);

	//==============================================================
	// Implementation of Methods Inherited from : ReferenceMaker
	//==============================================================
	RefResult NotifyRefChanged(Interval, RefTargetHandle, PartID&, RefMessage) {return REF_SUCCEED;}
	int				NumRefs				()		{ return 1; }
	RefTargetHandle GetReference		(int i) { return pblock; }
	void			SetReference		(int i, RefTargetHandle rtarg) { pblock=(IParamBlock2*)rtarg; }
	IOResult		Save				(ISave *isave);
	IOResult		Load				(ILoad *iload);


	void  GetValue( TimeValue t,  void *val, Interval &valid, GetSetMethod method=CTRL_ABSOLUTE);

	// for some stupid reason these two need to be implemented even though they don't work
	// and the MAX developers even use GetValue instead... stupid
	void  GetValueLocalTime (TimeValue t, void *val, Interval &valid, GetSetMethod method=CTRL_ABSOLUTE) {}
	void  SetValueLocalTime (TimeValue t, void *val, int commit=1, GetSetMethod method=CTRL_ABSOLUTE) {}

	// more stupid required implementations that would only make sense if this were a keyframe
	// controller
	void  Extrapolate(Interval range, TimeValue t, void *val, Interval &valid, int type) {}
	void*  CreateTempValue () { return NULL; }
	void  DeleteTempValue (void *val) {}
	void  ApplyValue (void *val, void *delta) {} 
	void  MultiplyValue (void *val, float m) {} 

	void  Copy(Control *from) {}

public:	// isntance data
	messiahData		messiah_data;

};

// ===============================================================
// messiahMAXcontrollerClassDesc
// ===============================================================
class messiahMAXcontrollerClassDesc : public ClassDesc2 
{
public:

	int 			IsPublic		()		{ return TRUE; }
	void *			Create			(BOOL loading = FALSE);
	const TCHAR *	ClassName		()		{ return GetString(IDS_CTL_CLASS_NAME); }
	SClass_ID		SuperClassID	()		{ return CTRL_MATRIX3_CLASS_ID; }
	Class_ID		ClassID			()		{ return MESSIAHCTRL_CLASS_ID; }
	const TCHAR* 	Category		()		{ return GetString(IDS_CATEGORY); }
	int				BeginCreate		(Interface *i);
	int				EndCreate		(Interface *i);

	// returns fixed parsable name (scripter-visible name)
	const TCHAR*	InternalName	()		{ return _T(MMAX_INTERNAL_NAME); }	
	
	// returns owning module handle
	HINSTANCE		HInstance		()		{ return hInstance; }				

};

// ===============================================================
// messiahDialogHook
// ===============================================================
class messiahDialogHook : public ParamMap2UserDlgProc
{
public:
	BOOL DlgProc(TimeValue t, IParamMap2 *map, HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
	void DeleteThis() {}
};

// ===============================================================
// objOnDeck
// ===============================================================
class objOnDeck
{
public:
	
	objOnDeck(int messiahObj, int cstFlags, int plugType, int plugFlags)
		: d_messiahObj(messiahObj), d_constructFlags(cstFlags), d_plugType(plugType), d_plugFlags(plugFlags)
	{}

	int		d_messiahObj;
	int		d_constructFlags;
	int		d_plugFlags;
	int		d_plugType;
};

// ===============================================================
// messiahMAXutility
// ===============================================================
class messiahMAXutility : public UtilityObj 
{
public:
	IUtil *		iu;
	Interface *	ip;
	HWND		hPanel;	
	HWND		hNumObjsLabel;
	HWND		hScale;
	HWND		hScaleGroup;
	HWND		hScaleLabel1;
	HWND		hScaleLabel2;
	HWND		hImportButton;
	float		d_scaleFactor;
	
	messiahMAXutility() : iu(0), ip(0), hPanel(0), hNumObjsLabel(0), 
						  d_scaleFactor(MMAX_INITIAL_SCALE) {}
	~messiahMAXutility();

	void BeginEditParams(Interface *ip,IUtil *iu);
	void EndEditParams(Interface *ip,IUtil *iu);
	void DeleteThis() {}

	void Init(HWND hWnd);		
	void ImportObjs();
	void Refresh();

	vector<objOnDeck *>	d_objQ;
};

static messiahMAXutility theUtilityObj;

// ===============================================================
// messiahMAXUtilClassDesc
// ===============================================================
class messiahMAXUtilClassDesc : public ClassDesc 
{
	public:
	int 			IsPublic()						{return 1;}
	void *			Create(BOOL loading = FALSE)	{return &theUtilityObj; MessageBox(NULL, "util created", NULL, 0); }
	const TCHAR *	ClassName()						{return GetString(IDS_UTIL_CLASS_NAME);}
	SClass_ID		SuperClassID()					{return UTILITY_CLASS_ID;}
	Class_ID		ClassID()						{return MESSIAHUTIL_CLASS_ID;}
	const TCHAR* 	Category()						{return _T("");}
};

static INT_PTR CALLBACK messiahMAXUtilDlgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam );

// ===============================================================
// mmax_exception - simple exception class
// ===============================================================
class mmax_exception
{
public:
	mmax_exception(char const *msg) { strncpy(d_msg, msg, 256); }
	char const *	what() const { return d_msg; }

private:

	char d_msg[256];
};

#endif