#include "MH_System.h"

#define HOST_MODULE

#include "messiahHOST.h"
#include <stdlib.h>

char _messiahDir[_MAX_PATH] = {0};

// ===============================================================
// Include platform specific implementations
// ===============================================================
#include "MH_SystemPlatforms.h"

// ===============================================================
// Initialize static variables
// ===============================================================
MH_System *	MH_System::_instance = 0;
void * (*MH_System::_hostGlobal)(char *) = 0;
double MH_System::scaleFactor = 1.0f;


// ===============================================================
// Constructor
// ===============================================================
MH_System::MH_System(char const *hostAppName, char const *hostModuleName) 
	:  _hostModuleName( hostModuleName ), _hostAppName( hostAppName )
{

}

// ===============================================================
// Destructor
// ===============================================================
MH_System::~MH_System()
{

}

// ===============================================================
// Retrieve Singleton instance and/or create platform specific object
// ===============================================================
MH_System *		MH_System::Inst(char const *hostAppName, char const *hostModuleName, void *(*hostGlobal)(char *) )
{
	if( !_instance )
	{
		if( !hostAppName || !hostModuleName )
			return 0;

		#ifdef _WIN32
		_instance = new MH_SystemWIN32(hostAppName, hostModuleName);
		#endif

		// other OS versions here

		// if we're in here then this is the creation of the instance
		// messiah initializations must be done here, not the constructor!
		_hostGlobal = hostGlobal;
		if( !_instance->_loadMessiahHost() || !_instance->_loadAllGlobals() || !_instance->_beginSession() )
		{
			delete _instance;
			_instance = 0;
		}

	}

	return _instance;
}

// ===============================================================
// Since the object is global it must be destroyed explicitly
// ===============================================================
void			MH_System::DestroyInst()
{
	if( _instance )
	{
		_instance->_endSession();
		_instance->_unloadMessiahHost();
		_instance->_unloadAllGlobals();

		delete _instance;
		_instance = 0;
	}
}

// ===============================================================
// Load all global messiahHOST variables
// ===============================================================
bool	MH_System::_loadAllGlobals()
{
	if( !messiahGlobal )
		return false;

	// System Globals //
	messiahStart		= ( int (*)(char[], char[], void*(*)(char *), int) )		messiahGlobal("messiahStart");
	messiah_current_frame = ( double * )									messiahGlobal("messiah_current_frame");
	messiahCheckTime	= ( void (*)(double, int) )							messiahGlobal("messiahCheckTime");
	messiah_Activated	= ( int * )											messiahGlobal("messiah_Activated");
	messiahInterface	= ( int (*)(void) )									messiahGlobal("messiahInterface");
	messiahShutdown		= ( void (*)(void) )								messiahGlobal("messiahShutdown");

	// Scene File Globals //
	messiahSceneFileCreateBuff	= ( char * (*)(void) )						messiahGlobal("messiahSceneFileCreateBuff");
	messiahSceneFileLoadBuff	= ( int (*)(char *) )						messiahGlobal("messiahSceneFileLoadBuff");
	messiahSceneLoad			= ( int (*)(int) )							messiahGlobal("messiahSceneLoad");
	messiahSceneClear			= ( void (*)(int) )							messiahGlobal("messiahSceneClear");
	messiahSceneFileFilterBuff	= ( int (*)(char **, char *) )				messiahGlobal("messiahSceneFileFilterBuff");
	messiahSceneFileClearBuff	= ( void (*)(char *) )						messiahGlobal("messiahSceneFileClearBuff");

	// Node Log Globals //
	num_nodes		= ( int * )												messiahGlobal("num_nodes");
	node_log		= ( FX_NodeLog * )										messiahGlobal("node_log");
	nodeLogElement	= ( FX_NodeLog * (*)(void *, void *, void *, int) )		messiahGlobal("nodeLogElement");
	nodeGetElement	= ( FX_NodeLog * (*)(void *, void *, void *, int) )		messiahGlobal("nodeGetElement");
	nodeGetFirst	= ( FX_NodeLog * (*)(void) )							messiahGlobal("nodeGetFirst");
	nodeGetLast		= ( FX_NodeLog * (*)(void) )							messiahGlobal("nodeGetLast");
	nodeSearchElement		= ( FX_NodeLog * (*)(int) )						messiahGlobal("nodeSearchElement");
	nodeClearElement		= ( int (*)(void *, int) )						messiahGlobal("nodeClearElement");
	nodeClearAllElements	= ( void (*)(int) )								messiahGlobal("nodeClearAllElements");

	// Node Object Globals //
	nodeObjectCreate		= ( int (*)(char *, int, int) )							messiahGlobal("nodeObjectCreate");
	nodeObjectRename		= ( int (*)(int, char *, int) )							messiahGlobal("nodeObjectRename");
	nodeObjectNameGet		= ( int (*)(int, char *, int) )							messiahGlobal("nodeObjectNameGet");
	nodeObjectCreatePoint	= ( int (*)(int, double, double, double) )				messiahGlobal("nodeObjectCreatePoint");
	nodeObjectGetPoint		= ( int (*)(int, int, double *, double *, double *) )	messiahGlobal("nodeObjectGetPoint");
	nodeObjectSetPoint		= ( int (*)(int, int, double, double, double) )			messiahGlobal("nodeObjectSetPoint");
	nodeObjectNumPoints		= (int (*)(int) )										messiahGlobal("nodeObjectNumPoints");
	nodeObjectCreateSurface	= ( int (*)(int, char *) )								messiahGlobal("nodeObjectCreateSurface");
	nodeObjectSurfaceSet	= ( void (*)(int, int, char *, void *) )				messiahGlobal("nodeObjectSurfaceSet");
	nodeObjectSurfaceGet	= ( int (*)(int, int, char *, void *) )					messiahGlobal("nodeObjectSurfaceGet");
	nodeObjectNumSurfaces	= ( int  (*)(int) )										messiahGlobal("nodeObjectNumSurfaces");
	nodeObjectCreatePoly	= ( void (*)(int, int, int *, unsigned short, int) )	messiahGlobal("nodeObjectCreatePoly");
	nodeObjectNumPolygons	= ( int	 (*)(int) )										messiahGlobal("nodeObjectNumPolygons");
	nodeObjectGetPolygon	= ( int	 (*)(int, int, int *, int, int *, int *) )		messiahGlobal("nodeObjectGetPolygon");
	nodeObjectAppend		= ( void (*)(int, int) )								messiahGlobal("nodeObjectAppend");
	nodeObjectEnd			= ( void (*)(int, int) )								messiahGlobal("nodeObjectEnd");
	nodeObjectRebuild		= ( int (*)(int, int) )									messiahGlobal("nodeObjectRebuild");

	// Motion
	nodeObjectGetMotion		= ( void (*)(int, double[TOTAL_NUM_CHANNELS], double) )	messiahGlobal("nodeObjectGetMotion");
	nodeObjectSetMotion		= ( void (*)(int, double[TOTAL_NUM_CHANNELS]) )			messiahGlobal("nodeObjectSetMotion");
	nodeObjectGetWorldPos	= ( void (*)(int, double[3], double) )					messiahGlobal("nodeObjectGetWorldPos");
	nodeObjectGetWorldRot	= ( void (*)(int, double[3], double) )					messiahGlobal("nodeObjectGetWorldRot");


	return true;
}

void MH_System::_unloadAllGlobals()
{
	messiahGlobal = 0;
	
	// System Globals //
	messiahStart			= 0;
	messiah_current_frame	= 0;
	messiahCheckTime		= 0;
	messiah_Activated		= 0;
	messiahInterface		= 0;
	messiahShutdown			= 0;

	// Scene File Globals //
	messiahSceneFileCreateBuff	= 0;
	messiahSceneFileLoadBuff	= 0;
	messiahSceneLoad			= 0;
	messiahSceneClear			= 0;
	messiahSceneFileFilterBuff	= 0;
	messiahSceneFileClearBuff	= 0;

	// Node Log Globals //
	num_nodes		= 0;
	node_log		= 0;
	nodeLogElement	= 0;
	nodeGetElement	= 0;
	nodeGetFirst	= 0;
	nodeGetLast		= 0;
	nodeSearchElement		= 0;
	nodeClearElement		= 0;
	nodeClearAllElements	= 0;

	// Node Object Globals //
	nodeObjectCreate		= 0;
	nodeObjectRename		= 0;
	nodeObjectCreatePoint	= 0;
	nodeObjectGetPoint		= 0;
	nodeObjectSetPoint		= 0;
	nodeObjectCreateSurface	= 0;
	nodeObjectSurfaceSet	= 0;
	nodeObjectCreatePoly	= 0;
	nodeObjectAppend		= 0;
	nodeObjectEnd			= 0;
	nodeObjectRebuild		= 0;

	// Motion
	nodeObjectGetMotion		= 0;
	nodeObjectSetMotion		= 0;
	nodeObjectGetWorldPos	= 0;
	nodeObjectGetWorldRot	= 0;
}

// ===============================================================
// messiah Start
// ===============================================================
int		MH_System::_beginSession()
{
	if( !messiahStart ) return 0;

	return messiahStart(const_cast<char *>( this->_hostAppName.c_str() ), 
						const_cast<char *>( this->_hostModuleName.c_str() ),
						_hostGlobal,
						MESSIAH_HOSTAPI_VERSION);
}

// ===============================================================
// messiah current frame
// ===============================================================
double	MH_System::CurrentFrame()
{
	if( !messiah_current_frame ) return 0.0;

	return *messiah_current_frame;
}

// ===============================================================
// messiah CheckTime
// ===============================================================
void	MH_System::CheckTime(double time, int flags)
{
	if( !messiahCheckTime ) return ;

	messiahCheckTime(time, flags);
}

// ===============================================================
// messiah Activated
// ===============================================================
int		MH_System::Activated()
{
	if( !messiah_Activated ) return 0;

	return *messiah_Activated;
}

// ===============================================================
// messiah Interface
// ===============================================================
int		MH_System::ShowInterface()
{
	if( !messiahInterface ) return 0;

	return messiahInterface();
}

// ===============================================================
// messiah Shutdown
// ===============================================================
void	MH_System::_endSession()
{
	if( !messiahShutdown ) return;

	messiahShutdown();
}