// ========			messiahDeformer.cpp			==============
//
// Copyright (C) 2002 pmG Worldwide LLC. All Rights Reserved.
//
// ===============================================================
//
// Description: 
//
// Implementation of messiahDeformer.
//
// ===============================================================
//
// Version History:
//
// Written By: Ron Griswold
// Date: 4/18/2002
//
// Revised By:
// Date: 
//
// ===============================================================

#include "messiahDeformerNode.h"

#include <maya/MFnUnitAttribute.h>
#include <maya/MFnNumericAttribute.h>
#include <maya/MFnMessageAttribute.h>
#include <maya/MFnTypedAttribute.h>

#include <maya/MTime.h>
#include <maya/MDagPath.h>
#include <maya/MGlobal.h>
#include <maya/MTypeId.h>
#include <maya/MPoint.h>
#include <maya/MMatrix.h>

#include <maya/MItGeometry.h>

#include <maya/MPlug.h>
#include <maya/MPlugArray.h>
#include <maya/MDataBlock.h>
#include <maya/MDataHandle.h>

#include <maya/MFnDagNode.h>
#include <maya/MFnDependencyNode.h>

// Get rid of this:
//#define MESSIAH_HOST_MODULE
//#include "messiahHOST.h"


// ===============================================================
// Initialize node type id
// ===============================================================
MTypeId		messiahDeformer::id( MM_DISP_ID );

// ===============================================================
// Initialize attribs to NULL
// ===============================================================
MObject     messiahDeformer::time;       
MObject		messiahDeformer::hostItem;
MObject		messiahDeformer::loadId;

// ===============================================================
// Constructor
// ===============================================================
messiahDeformer::messiahDeformer() //: _connected( false )
{
	// Add the node to the list of nodes in the manager
	// See messiahMayaManager.h
	messiahMayaManager::Inst()->AddNode( this );
}

// ===============================================================
// Destructor
// ===============================================================
messiahDeformer::~messiahDeformer()
{
	// Remove this node from the list of nodes in the manager
	// See messiahMayaManager.h
	messiahMayaManager::Inst()->RemoveNode( this );
}

// ===============================================================
// Creator
// ===============================================================
void * messiahDeformer::creator()
{
	return new messiahDeformer();
}

// ===============================================================
// Initialize
// ===============================================================
MStatus messiahDeformer::initialize()
{
	// initialize the attributes of the node
	MFnNumericAttribute		nAttr;
	MFnUnitAttribute		unitAttr;
	MFnMessageAttribute		mAttr;

	MStatus				status;

	hostItem = mAttr.create( MM_ATR_HOST_ITEM_L, MM_ATR_HOST_ITEM_S, &status);
	mAttr.setReadable(false);

	loadId = nAttr.create( MM_ATR_LOAD_ID_L, MM_ATR_LOAD_ID_S, MFnNumericData::kLong );
	nAttr.setStorable( true );
	nAttr.setHidden( true );

	messiahDeformer::time = unitAttr.create(MM_ATR_TIME_L, MM_ATR_TIME_S, MFnUnitAttribute::kTime, 0.0, &status);
	unitAttr.setReadable( false );
	unitAttr.setKeyable( false );
	unitAttr.setCached( false );
	MM_CheckError(status, "Failed to allocate time attribute");

	status = addAttribute( hostItem );
		if (!status) { status.perror("addAttribute"); return status;}

	status = addAttribute( time );
		if (!status) { status.perror("addAttribute"); return status;}

	status = addAttribute( loadId );
		if (!status) { status.perror("addAttribute"); return status;}

	attributeAffects(time, outputGeom);

	return status;
}

// ===============================================================
// Deform
// ===============================================================
MStatus messiahDeformer::deform(MDataBlock &block, MItGeometry &iter, MMatrix const &mat, unsigned int multiIndex)
{
	MStatus status = MS::kSuccess;

	MH_NodeObject	no( this->_nodeLog->obj );
	double			x, y, z, i;

	x = y = z = i = 0;

	// iterate through each point in the geometry
	for(iter.reset(), i = 0; !iter.isDone(); iter.next(), i++ ) 
	{
		no.GetPoint(i, &x, &y, &z);
		
		MPoint pt( -x/MHS::scaleFactor, y/MHS::scaleFactor, z/MHS::scaleFactor );

		iter.setPosition(pt);
	}
	return status;
}

extern int temporary_messiahobj_global;	//messiahCommand.cpp

// ===============================================================
// Connection Made
// ===============================================================
MStatus	messiahDeformer::connectionMade(const MPlug &plug, const MPlug &otherPlug, bool asSrc )
{

	MStatus stat;

	// ===============================================================
	// plug was connected to hostItem
	// ===============================================================
	if( plug == hostItem )
	{
		// Get the name of the host item ////////////////////////
		MObject			SourceObject = otherPlug.node();
		MFnDagNode		SourceDagNode( SourceObject );
		MDagPath		SourceDagPath;

		SourceDagNode.getPath( SourceDagPath );
		MString			PathName = SourceDagPath.fullPathName();
		char const *	tmp = PathName.asChar();
		char *			SourceName = new char[::strlen(tmp)+1];
		::strcpy( SourceName, tmp );
		
		MakeValidMessiahName( SourceName );
		/////////////////////////////////////////////////////////

		// Create the FX_NodeLog ////////////////////////////////
		this->_nodeLog = MH_NodeLogFuncs::LogElement(NULL, (void *)this, NULL, NULL);
		this->_nodeLog->type = FX_NODE_DISPLACEMENT;
		/////////////////////////////////////////////////////////

		// Add item to messiah //////////////////////////////////
		if( ! (MAN::Inst()->SceneState() & MAN::SceneLoading) )
		{
			if( !temporary_messiahobj_global )
			{
			this->_nodeLog->obj = (MH_NodeObject::Create( SourceName, 0, 0 ))->MessiahObj();
			common::hostObjectGeometryCopy( this->_nodeLog, 1 );
		}
		else
		{
				this->_nodeLog->obj = temporary_messiahobj_global;
				temporary_messiahobj_global = 0;
			}
		}
		else
		{
			MPlug				loadId;
			int					id;
			MFnDependencyNode	dNode( this->thisMObject() );

			loadId = dNode.findPlug(MM_ATR_LOAD_ID_L, &stat);
			if(!stat)
			{
				MGlobal::displayError("Error Loading messiahDeformer");
				return MS::kFailure;
			}

			loadId.getValue( id );
			this->_nodeLog->load_ID = id;
		}
		/////////////////////////////////////////////////////////

	}// plug == hostItem

	return MS::kUnknownParameter;

}

// ===============================================================
// Compute - subclasses MPxDeformerNode::compute()
// Needed to get notification when input geom changes
// ===============================================================
MStatus messiahDeformer::compute(MPlug const &plug, MDataBlock &dataBlock)
{

	// we have to call this even thought we are relying on the global time provided 
	// by the DG callback "TimeChanged". 
	// See messiahMayaManager::OnTimeChange
	MDataHandle frameH = dataBlock.inputValue( time );

	return MS::kUnknownParameter;
}
