/////////////////////////////////////////////////////////////
// CINEMA SDK : CINEMA OPERATING SYSTEM        						 //
/////////////////////////////////////////////////////////////
// VERSION    : CINEMA 4D V6 AMBER												 //
/////////////////////////////////////////////////////////////
// (c) 1989-2001 MAXON Computer GmbH, all rights reserved	 //
/////////////////////////////////////////////////////////////

// C interface functions

#ifndef __OPERATINGSYSTEM_H
#define __OPERATINGSYSTEM_H

#include "ge_math.h"
#include "ge_prepass.h"
#include "c4d_basetime.h"

class  RootTextureString;
class  Registry;
class  CDialog;
class  CUserArea;
class  _GeListView;
class  _SimpleListView;
class  BaseTag;
class  MPBaseThread;
class  Semaphore;
class  VariableTag;
class  BaseTime;
class	 BaseChannel;
class	 BaseContainer;
class	 BaseDocument;
class	 BaseSelect;
class  HyperFile;
class  BaseList2D;
class  BaseList4D;
class  BaseObject;
class  SplineObject;
class  PointObject;
class  PolygonObject;
class  LineObject;
class  PluginObject;
class  BaseDraw;
class  BaseDrawHelp;
class  BaseView;
class  BaseTrack;
class  String;
class  BaseBitmap;
class  MemoryPool;
class  BaseMaterial;
class  Material;
class  PluginMaterial;
class	 RenderData;
class  Render;
class	 TextureTag;
class  MovieSaver;
class  BrowseFiles;
class  Parser;
class  BaseFile;
class  SelectionTag;
class  CPluginTag;
class  UVWTag;
class  ObjectSafety;
class  ParticleObject;
class  PluginObject;
class  ParticleTag;
class  StickTextureTag;
class  Particle;
class  LocalResource;
class  HierarchyHelp;
class  BaseKey;
class  BaseSequence;
class  BaseTrack;
class  PSRKey;
class  PLAKey;
class  MarkerKey;
class  FolderSequence;
class  SoundSequence;
class  BaseSound;
class	 BaseThread;
class  EnumerateEdges;
class  PaintTexture;
class  UVClass;
class  EditorWindow;
class  VPBuffer;
class  CPluginKey;
class  CPluginSequence;
class  CPluginTrack;
class  GeData;
class  _TextureGroup;
class  _BaseContainer;
class  _Filename;
class  _String;

struct C4D_Painter;
struct C4D_BitmapFilter;
struct ParticleDetails;
struct PolyInfo;
struct SData;
struct VolumeData;
struct Ray;
struct Tangent;
struct UVWStruct;
struct Segment;
struct Polygon;
struct TimingStruct;
struct LVector;
struct LMatrix;
struct InitRenderStruct;
struct RayObject;
struct RayLight;
struct PolyVector;
struct TexData;
struct VideoPostStruct;
struct SurfaceData;
struct PMStruct;
struct PLUGIN_DATA;
struct VIDEOPOST;
struct CHANNELSHADER;
struct VOLUMESHADER;
struct VP_DATA;
struct MouseDownInfo;
struct DrawInfo;
struct RayLightCache;
struct IconData;

#define API_VERSION	7300

#define C4DPL_SHADER2D			 2
#define C4DPL_SHADER3D			 3
#define C4DPL_FILTER2D			 5
#define C4DPL_FILTER3D			 6
#define C4DPL_MENU					 7
#define C4DPL_OBJECT				 8
#define C4DPL_ANIMATION			 9
#define C4DPL_TAG						10
#define C4DPL_BITMAPFILTER	12
#define C4DPL_VIDEOPOST			13
#define C4DPL_TOOL					14
#define C4DPL_SCENEHOOK			15

enum
{
	DIALOG_PIN = 1,						// LONG flags
	DIALOG_CHECKBOX,					// LONG id, LONG flags, String *name
	DIALOG_STATICTEXT,				// LONG id, LONG flags, String *name, LONG borderstyle
	DIALOG_BUTTON,						// LONG id, LONG flags, String *name
	DIALOG_ARROWBUTTON,				// LONG id, LONG flags, LONG arrowtype
	DIALOG_EDITTEXT,					// LONG id, LONG flags
	DIALOG_EDITNUMBER,				// LONG id, LONG flags
	DIALOG_EDITNUMBERUD,			// LONG id, LONG flags
	DIALOG_EDITSLIDER,				// LONG id, LONG flags
	DIALOG_SLIDER,						// LONG id, LONG flags
	DIALOG_COLORFIELD,				// LONG id, LONG flags
	DIALOG_COLORCHOOSER,			// LONG id, LONG flags
	DIALOG_USERAREA,					// LONG id, LONG flags
	DIALOG_RADIOGROUP,				// LONG id, LONG flags
	DIALOG_COMBOBOX,					// LONG id, LONG flags
	DIALOG_POPUPBUTTON,				// LONG id, LONG flags
	DIALOG_CHILD,							// LONG id, LONG subid, String *child);
	DIALOG_FREECHILDS,				// LONG id
	DIALOG_DLGGROUP,					// LONG flags
	DIALOG_SETTITLE,					// String *name
	DIALOG_GROUPSPACE,				// spacex,spacey
	DIALOG_GROUPBORDER,				// borderstyle
	DIALOG_GROUPBORDERSIZE,	  // left, top, right, bottom
	DIALOG_SETIDS,						// left, top, right, bottom
	DIALOG_LAYOUTCHANGED,			// relayout dialog
	DIALOG_ACTIVATE,					// activate gadget
	DIALOG_ADDSUBMENU,				// add submenu, text
	DIALOG_ENDSUBMENU,				// add submenu, text
	DIALOG_ADDMENUCMD,				// add menutext, text, cmdid
	DIALOG_FLUSHMENU,					// delete all menu entries
	DIALOG_INIT,							// create new layout
	DIALOG_CHECKNUMBERS,			// range check of all dialog elements
	DIALOG_SETGROUP,					// set group as insert group
	DIALOG_FLUSHGROUP,				// flush all elements of this group and set insert point to this group
	DIALOG_SETMENU,						// set and display menu in manager
	DIALOG_SCREEN2LOCALX,			// Screen2Local coordinates
	DIALOG_SCREEN2LOCALY,			// Screen2Local coordinates
	DIALOG_ADDMENUSTR,				// add menutext, text, id
	DIALOG_RADIOBUTTON,				// LONG id, LONG flags, String *name
	DIALOG_ADDMENUSEP,				// add menu separator
	DIALOG_SEPARATOR,					// add separator
	DIALOG_MULTILINEEDITTEXT, // add multiline editfield
	DIALOG_INITMENUSTR,				// enable/disable/check/uncheck menustrings
	DIALOG_RADIOTEXT,
	DIALOG_MENURESOURCE,			// private for painter
	DIALOG_LISTVIEW,					// LONG id, LONG flags
	DIALOG_SUBDIALOG,					// LONG id, LONG flags
	DIALOG_CHECKCLOSE,				// CheckCloseMessage for dialog
	DIALOG_GETTRISTATE,				// get TriState-flag of the specified gadget

	DIALOG_
};

enum
{
	LV_GETLINECOUNT    = 1,  // request the number of lines of the listview
	LV_GETCOLUMNCOUNT  = 2,  // request the number of columns of listview
	LV_GETLINEHEIGHT   = 3,  // ask for the line height of the specific 'line'
	LV_GETCOLUMNWIDTH  = 4,  // ask for the width of the specific 'column' in 'line'
	LV_GETCOLUMTYPE    = 5,  // ask for the type of the column in 'line',
		LV_COLUMN_TEXT     = 'txt',
		LV_COLUMN_EDITTEXT = 'edt',
		LV_COLUMN_BMP      = 'bmp',
		LV_COLUMN_CHECKBOX = 'chk',
		LV_COLUMN_BUTTON   = 'btn',
		LV_COLUMN_USERDRAW = 'usr',

	LV_GETCOLUMDATA    = 6,  // ask for the data of the column in 'line',
	LV_GETLINESELECTED = 7,  // ask if the line is selected
	LV_GETCOLSPACE     = 8,  // ask for space in pixels between two columns
	LV_GETLINESPACE    = 9,  // ask for space in pixels between two lines
	LV_GETFIXEDLAYOUT  = 10, // ask for fixed layout: FALSE...indiv. layout, TRUE...fixed layout
	LV_DESTROYLISTVIEW = 11, // destroy listview and all data
	LV_INITCACHE       = 12, // (internal) before call the listview

	LV_LMOUSEDOWN       = 50, // mouse down at line, col,
	LV_ACTION						= 51, // gadget command, col, data1 = msg,
	LV_USERDRAW					= 52,
	LV_REDRAW						= 53, // redraw the listview (supermessage)
	LV_DATACHANGED		  = 54, // layout data has changed
	LV_SHOWLINE					= 55, // scroll line into the visible area

	LV_SIMPLE_SELECTIONCHANGED = 100, // simplelistview: selection changed
	LV_SIMPLE_CHECKBOXCHANGED  = 101, // simplelistview: checkbox changed
	LV_SIMPLE_FOCUSITEM        = 102, // simplelistview: set focus to item
	LV_SIMPLE_BUTTONCLICK      = 103, // simplelistview: button click
		LV_SIMPLE_ITEM_ID	= 1,
		LV_SIMPLE_COL_ID	= 2,
		LV_SIMPLE_DATA    = 3,
	LV_SIMPLE_DOUBLECLICK      = 104, // doubleclick occured

	// result types
	LV_RES_LONG        = 'long',
	LV_RES_BITMAP      = 'bmp',
	LV_RES_STRING      = 'strg',
	LV_RES_NIL         = 'nil',

	LV__
};

enum
{
	DLG_OK     = 1,
	DLG_CANCEL = 2,
	DLG_HELP   = 4,
	DLG_
};

enum StringEncoding
{
	StXbit		= 0,
	St8bit		= 1,
	St7bit		= 2,
	St7bithex	= 3
};

#define CLONE_IDENTMARKER  				(1<<0) // tags, objects and material
#define CLONE_ISOPARM							(1<<1) // only objects
#define CLONE_NO_HIERARCHY				(1<<2)
#define CLONE_NO_ANIMATION				(1<<3)
#define CLONE_NO_BITS							(1<<4)

#define TAG_VISIBLE								(1<<0)
#define TAG_MULTIPLE							(1<<1)
#define TAG_HIERARCHICAL					(1<<2)
#define TAG_EXPRESSION						(1<<3)

#define OBJECT_MODIFIER						(1<<0)
#define OBJECT_HIERARCHYMODIFIER	(1<<1)
#define OBJECT_GENERATOR					(1<<2)
#define OBJECT_INPUT							(1<<3)
#define OBJECT_LODDEPENDENT				(1<<4)

#define SETTINGS_GENERAL							0
#define SETTINGS_MODELER							1
#define SETTINGS_DOCUMENT							2
#define SETTINGS_OBJECTMANAGER				3
#define SETTINGS_MATERIALMANAGER			4
#define SETTINGS_COORDINATEMANAGER		5
#define SETTINGS_STRUCTUREMANAGER			6
#define SETTINGS_TIMELINE							7

// Material::Update
#define MATERIALUPDATE_NOPREVIEW			0
#define MATERIALUPDATE_PREVIEW				1
#define MATERIALUPDATE_FASTPREVIEW		2

// GeGetSerialInfo
#define SERIAL_CINEMA4D		0
#define SERIAL_BODYPAINT	1

// GeGetVersionType
#define VERSION_C4DXL			(1<<0)
#define VERSION_BODYPAINT	(1<<1)
#define VERSION_NET				(1<<2)
#define VERSION_SERVER		(1<<3)
#define VERSION_DEMO			(1<<4)
#define VERSION_C4DART    (1<<5)

// file selector type
#define FSTYPE_ANYTHING								0
#define FSTYPE_IMAGES									1
#define FSTYPE_SCENES									2
#define FSTYPE_COFFEE									3

// document IDs
#define DOCUMENT_TIME										1000 // BASETIME
#define DOCUMENT_FPS										1001 // LONG
#define DOCUMENT_MINTIME								1002 // BASETIME
#define DOCUMENT_MAXTIME								1003 // BASETIME
#define DOCUMENT_LOOPMINTIME						1004 // BASETIME
#define DOCUMENT_LOOPMAXTIME						1005 // BASETIME
#define DOCUMENT_LOD										1006 // REAL
#define DOCUMENT_RENDERLOD							1007 // BOOL
#define DOCUMENT_USEANIMATION						1008 // BOOL
#define DOCUMENT_USEEXPRESSIONS					1009 // BOOL
#define DOCUMENT_USEGENERATORS					1010 // BOOL
#define DOCUMENT_USEDEFORMERS						1011 // BOOL
#define DOCUMENT_PATH										1012 // FILENAME
#define DOCUMENT_NAME										1013 // FILENAME
#define DOCUMENT_STATEX									1014 // BOOL
#define DOCUMENT_STATEY									1015 // BOOL
#define DOCUMENT_STATEZ									1016 // BOOL
#define DOCUMENT_STATEW									1017 // BOOL
#define DOCUMENT_CHANGED								1018 // BOOL
#define DOCUMENT_MODE										1019 // LONG
#define DOCUMENT_ACTION									1020 // LONG

// tag IDs
#define COMPOSITINGTAG_CASTSHADOW				1000 // BOOL
#define COMPOSITINGTAG_RECEIVESHADOW		1001 // BOOL
#define COMPOSITINGTAG_SEENBYCAMERA			1002 // BOOL
#define COMPOSITINGTAG_SEENBYRAYS				1003 // BOOL
#define COMPOSITINGTAG_BACKGROUND				1004 // BOOL
#define COMPOSITINGTAG_AAOVERWRITE			1005 // BOOL
#define COMPOSITINGTAG_AAMINLEVEL				1006 // LONG
#define COMPOSITINGTAG_AAMAXLEVEL				1007 // LONG
#define COMPOSITINGTAG_GI               1008 // BOOL
#define COMPOSITINGTAG_GIACCURACY				1009 // REAL
#define COMPOSITINGTAG_ENABLECHN0				1010 // BOOL
#define COMPOSITINGTAG_ENABLECHN1				1011 // BOOL
#define COMPOSITINGTAG_ENABLECHN2				1012 // BOOL
#define COMPOSITINGTAG_ENABLECHN3				1013 // BOOL
#define COMPOSITINGTAG_ENABLECHN4				1014 // BOOL
#define COMPOSITINGTAG_ENABLECHN5				1015 // BOOL
#define COMPOSITINGTAG_AATHRESHOLD			1016 // REAL
#define COMPOSITINGTAG_SEENBYGI					1017 // BOOL
#define COMPOSITINGTAG_IDCHN0						1110 // LONG
#define COMPOSITINGTAG_IDCHN1						1111 // LONG
#define COMPOSITINGTAG_IDCHN2						1112 // LONG
#define COMPOSITINGTAG_IDCHN3						1113 // LONG
#define COMPOSITINGTAG_IDCHN4						1114 // LONG
#define COMPOSITINGTAG_IDCHN5						1115 // LONG

#define VERTEXMAPTAG_NAME								1000 // STRING

#define METABALLTAG_NEGATIVE						1000 // BOOL
#define METABALLTAG_STRENGTH						1001 // REAL
#define METABALLTAG_RADIUS							1002 // REAL

#define POINTSELECTIONTAG_NAME					1000 // STRING

#define POLYSELECTIONTAG_NAME						1000 // STRING

#define KINEMATICTAG_WX									1000 // BOOL
#define KINEMATICTAG_WY									1001 // BOOL
#define KINEMATICTAG_WZ									1002 // BOOL
#define KINEMATICTAG_MIN								1003 // VECTOR
#define KINEMATICTAG_MAX								1004 // VECTOR
#define KINEMATICTAG_DAMPING						1005 // REAL

#define DISPLAYTAG_DISPLAYMODE					1000 // LONG
#define DISPLAYTAG_BACKFACECULLING			1001 // BOOL
#define DISPLAYTAG_TEXTURES							1002 // BOOL
#define DISPLAYTAG_LOD									1003 // BOOL
#define DISPLAYTAG_LODVALUE							1004 // REAL

#define TEXTURETAG_OFFSETX							1000 // REAL
#define TEXTURETAG_OFFSETY							1001 // REAL
#define TEXTURETAG_LENGTHX							1002 // REAL
#define TEXTURETAG_LENGTHY							1003 // REAL
#define TEXTURETAG_PROJECTION						1004 // LONG
#define TEXTURETAG_SIDE									1005 // LONG
#define TEXTURETAG_RESTRICTION					1006 // STRING
#define TEXTURETAG_TILE									1007 // BOOL
#define TEXTURETAG_SEAMLESS							1008 // BOOL
#define TEXTURETAG_MIX									1009 // BOOL

#define STICKTEXTURETAG_ACTIVE					1000 // BOOL

// get/set mode flags
#define MODE_EDITOR	0
#define MODE_RENDER	1
#define MODE_DEFORM	2

// maximum number of texture paths
#define MAX_GLOBAL_TEXTURE_PATHS	10

// basedraw data
#define BASEDRAW_DISABLETEXTURES				1000 // BOOL
#define BASEDRAW_DISABLEBACKCULL				1001 // BOOL
#define BASEDRAW_SHOWSAFEFRAME					1002 // BOOL
#define BASEDRAW_DISPLAYACTIVE					1003 // LONG
#define BASEDRAW_DISPLAYINACTIVE				1004 // LONG
#define BASEDRAW_USEPROPERTIESACTIVE		1005 // BOOL
#define BASEDRAW_USEPROPERTIESINACTIVE	1006 // BOOL
#define BASEDRAW_SHOWNORMALS						1007 // BOOL
#define BASEDRAW_SHOWPATH								1008 // BOOL
#define BASEDRAW_SHOWPICTURE						1009 // BOOL
#define BASEDRAW_XRAY										1010 // BOOL
#define BASEDRAW_PICTURE								1011 // FILENAME
#define BASEDRAW_OFFSETX								1012 // REAL
#define BASEDRAW_OFFSETY								1013 // REAL
#define BASEDRAW_SIZEX									1014 // REAL
#define BASEDRAW_SIZEY									1015 // REAL

#define PAINTER_SAVETEXTURE							1000 // save textures (if doc -> all doc associated, if -> tex only tex, otherwise all textures)

#define OPTIMALMAPPING_PRESERVEORIENTATION	1000	// BOOL
#define OPTIMALMAPPING_STRETCHTOFIT					1001	// BOOL
#define OPTIMALMAPPING_DISTORTION						1002	// REAL
#define OPTIMALMAPPING_SPACING							1003	// REAL

typedef LONG CDialogMessage(CDialog *cd, CUserArea *cu, _BaseContainer *msg);
typedef void ListViewCallBack(LONG &res_type,void *&result,void *userdata,void *secret,LONG cmd,LONG line,LONG col,void *data1);
typedef void IlluminationModel(VolumeData *sd, RayLightCache *rlc, void *dat);

typedef void ThreadMain(void *data);
typedef Bool ThreadTest(void *data);
typedef void ProgressHook(Real p, void *private_data);

typedef void *HierarchyAlloc(void *main);
typedef void HierarchyFree(void *main, void *data);
typedef void HierarchyCopyTo(void *main, void *src, void *dst);
typedef Bool HierarchyDo(void *main, void *data, BaseObject *op, const Matrix &mg, Bool controlobject);

#define VPGETINFO_XRESOLUTION	0
#define VPGETINFO_YRESOLUTION	1
#define VPGETINFO_BITDEPTH		2
#define VPGETINFO_CPP					3
#define VPGETINFO_VISIBLE			4
#define VPGETINFO_LINEOFFSET	5 // offset of component in line

struct C4D_Shader
{
	Real						(*SNoise							)(Vector p);
	Real						(*SNoiseT							)(Vector p, Real t);
	Real						(*Noise								)(Vector p);
	Real						(*NoiseT							)(Vector p, Real t);
	Real						(*PNoise							)(Vector p, Vector d);
	Real						(*PNoiseT							)(Vector p, Real t, Vector d, Real dt);
	Real						(*Turbulence					)(Vector p, Real oct, Bool abs);
	Real						(*TurbulenceT					)(Vector p, Real t, Real oct, Bool abs);
	Real						(*WavyTurbulence			)(Vector p, Real t, Real oct, Real start);
	void						(*InitFbm							)(Real *table, LONG max_octaves, Real lacunarity, Real h);
	Real						(*Fbm									)(Real *table, Vector p, Real oct);
	Real						(*FbmT								)(Real *table, Vector p, Real t, Real oct);
	Real						(*RidgedMultifractal	)(Real *table, Vector p, Real oct, Real offset, Real gain);
	Real						(*CalcSpline					)(Real x, Real *knot, LONG nknots);
	Vector					(*CalcSplineV					)(Real x, Vector *knot, LONG nknots);

	void						(*Illuminance1				)(VolumeData *sd, Vector *diffuse, Vector *specular, Real exponent);
	void						(*Illuminance2				)(VolumeData *sd, Vector *diffuse, Vector *specular, Real exponent, Vector *p, Vector *n, Vector *bumpn, Vector *rayv, LONG calc_shadow);
	Vector					(*Illuminance3				)(VolumeData *sd, Vector *p, LONG calc_shadow);
	Bool						(*IlluminateEx        )(VolumeData *sd, RayLight *rl, Vector *color, Vector *light_vector, const Vector &p, const Vector &n, LONG calc_shadow, LONG lhit);
	void						(*LightCalcSpecial    )(VolumeData *sd, RayLight *rl, const Vector &col, Real exponent, const Vector &rayv, Real cosc, const Vector &p, const Vector &bumpn, Vector *diffuse, Vector *specular);

	LONG						(*GetCurrentCPU				)(VolumeData *sd);
	LONG						(*GetCPUCount					)(VolumeData *sd);
	void*						(*GetRayData					)(VolumeData *sd, LONG i);
	Vector					(*TraceColor					)(VolumeData *sd, Ray *ray, Real maxdist, LONG lhit, Vector *p, Vector *n, LONG *hit);
	void  					(*TraceGeometry				)(VolumeData *sd, Ray *ray, Real maxdist, LONG lhit, Vector *p, Vector *n, LONG *hit);

	RayObject*			(*GetObj							)(VolumeData *sd, LONG i);
	LONG						(*GetObjCount					)(VolumeData *sd);

	RayLight*				(*GetLight						)(VolumeData *sd, LONG i);
	LONG						(*GetLightCount				)(VolumeData *sd);

	RayObject*			(*ID_to_Obj						)(VolumeData *sd, LONG id, LONG *vind);
	LONG						(*Obj_to_ID						)(VolumeData *sd, RayObject *op);

	void						(*GetUVW							)(VolumeData *sd, RayObject *op, LONG uvwind, LONG i, PolyVector *uvw);
	Vector					(*GetPointUVW					)(VolumeData *sd, TexData *tdp, LONG lhit, const Vector &p);
	void						(*GetNormals					)(VolumeData *sd, RayObject *op, LONG i, PolyVector *norm);
	TexData*				(*GetTexData					)(VolumeData *sd, RayObject *op, LONG i);
	Vector					(*GetNormal						)(VolumeData *sd, RayObject *op, LONG i);
	Vector					(*GetSmoothedNormal		)(VolumeData *sd, LONG id, const Vector &p);
	Bool						(*GetRS								)(VolumeData *sd, LONG id, const Vector &p, Real *r, Real *s);

	void						(*GetRay							)(VolumeData *sd, Real x, Real y, Ray *ray);
	Vector					(*ScreenToCamera			)(VolumeData *sd, const Vector &p);
	Vector					(*CameraToScreen			)(VolumeData *sd, const Vector &p);
	Bool						(*ProjectPoint				)(VolumeData *sd, TexData *tdp, LONG lhit, const Vector &p, const Vector &n, Vector *uv);

	Vector					(*CalcHardShadow			)(VolumeData *sd, const Vector &p1, const Vector &p2, Bool transparency, LONG lhit);
	Vector					(*CalcShadow					)(VolumeData *sd, RayLight *l, const Vector &p, const Vector &n, Bool transparency, LONG lhit);

	VolumeData*			(*AllocVolumeData			)(void);
	void						(*FreeVolumeData			)(VolumeData *sd);

	Bool						(*Illuminate					)(VolumeData *sd, RayLight *rl, Vector *color, Vector *light_vector, const Vector &p, const Vector &n, LONG calc_shadow, LONG lhit, Vector *xshadow);

	void						(*StatusSetText				)(VolumeData *sd, _String *str);
	void						(*StatusSetBar				)(VolumeData *sd, Real percentage);

	TexData*				(*AllocTexData				)(void);
	void						(*FreeTexData					)(TexData *td);
	void						(*InitTexData					)(TexData *td);

	Vector					(*CalcVisibleLight		)(VolumeData *sd, Ray *ray, Real maxdist, Vector *trans);
	void						(*GetDUDV							)(VolumeData *vd, TexData *tex, const Vector &p, const Vector &n, Vector *ddu, Vector *ddv);
	void						(*GetXY								)(VolumeData *sd, LONG *x, LONG *y, LONG *scale);

	LONG						(*Obj_to_Num					)(VolumeData *sd, RayObject *obj);
	LONG						(*Light_to_Num				)(VolumeData *sd, RayLight *light);

	Bool						(*GetVideopost				)(VolumeData *sd, LONG i, LONG *id, VIDEOPOST **vp, VP_DATA **data);
	LONG						(*GetVideopostCount		)(VolumeData *sd);

	void						(*CopyVolumeData			)(VolumeData *src, VolumeData *dst);
	Vector					(*TraceColorDirect		)(VolumeData *sd, Ray *ray, Real maxdist, LONG raydepth, LONG raybits, LONG lhit, LVector *p, LVector *n, LONG *hit);

	Bool						(*VPAllocateBuffer		)(Render *render, LONG id, LONG subid, LONG bitdepth, Bool visible);
	VPBuffer*				(*VPGetBuffer					)(Render *render, LONG id, LONG subid);
	LONG						(*VPGetInfo						)(VPBuffer *buf, LONG type);
	Bool						(*VPGetLine						)(VPBuffer *buf, LONG x, LONG y, LONG cnt, void *data, LONG bitdepth, Bool dithering);
	Bool						(*VPSetLine						)(VPBuffer *buf, LONG x, LONG y, LONG cnt, void *data, LONG bitdepth, Bool dithering);
	LONG						(*VPSave							)(VPBuffer *buf, _Filename *name, LONG format, _BaseContainer *data, LONG savebits);

	void						(*OutOfMemory					)(VolumeData *sd);
	void  					(*GetSurfaceData			)(VolumeData *sd, SurfaceData *cd, Bool calc_illum, Bool calc_shadow, Bool calc_refl, Bool calc_trans, Ray *ray, const Vector &p, const Vector &n, LONG lhit);
	Real						(*GetLightFalloff			)(LONG falloff, Real inner, Real outer, Real dist);

	Bool						(*TestBreak						)(VolumeData *sd);

	void						(*VPRegisterBuffer		)(Render *render, LONG id, LONG bitdepth, LONG cpp, _String *name);
	void						(*VPUpdateLine				)(Render *render, LONG line);
	void						(*VPUpdateScreen			)(Render *render);
	void						(*VPGetRenderData			)(Render *render, _BaseContainer *bc);
	void						(*VPSetRenderData			)(Render *render, _BaseContainer *bc);

	void						(*Illuminance					)(VolumeData *sd, Vector *diffuse, Vector *specular, IlluminationModel *model, void *data);
};

struct C4D_HyperFile
{
	Bool						(*WriteChar						)(HyperFile *hf, CHAR  v);
	Bool						(*WriteUChar					)(HyperFile *hf, UCHAR v);
	Bool						(*WriteWord						)(HyperFile *hf, WORD  v);
	Bool						(*WriteUWord					)(HyperFile *hf, UWORD v);
	Bool						(*WriteLong						)(HyperFile *hf, LONG  v);
	Bool						(*WriteULong					)(HyperFile *hf, ULONG v);
	Bool						(*WriteReal						)(HyperFile *hf, Real  v);
	Bool						(*WriteLReal					)(HyperFile *hf, LReal v);
	Bool						(*WriteBool						)(HyperFile *hf, Bool  v);
	Bool						(*WriteTime						)(HyperFile *hf, const BaseTime &v);
	Bool						(*WriteVector					)(HyperFile *hf, const Vector &v);
	Bool						(*WriteLVector				)(HyperFile *hf, const LVector &v);
	Bool						(*WriteMatrix					)(HyperFile *hf, const Matrix &v);
	Bool						(*WriteLMatrix				)(HyperFile *hf, const LMatrix &v);
	Bool						(*WriteString					)(HyperFile *hf, _String *v);
	Bool						(*WriteFilename				)(HyperFile *hf, _Filename *v);
	Bool						(*WriteImage					)(HyperFile *hf, BaseBitmap *v, LONG format, _BaseContainer *data);
	Bool						(*WriteContainer			)(HyperFile *hf, _BaseContainer *v);
	Bool						(*WriteMemory					)(HyperFile *hf, void *data, LONG count);

	Bool						(*ReadChar						)(HyperFile *hf, CHAR          *v);
	Bool						(*ReadUChar						)(HyperFile *hf, UCHAR         *v);
	Bool						(*ReadWord						)(HyperFile *hf, WORD          *v);
	Bool						(*ReadUWord					  )(HyperFile *hf, UWORD         *v);
	Bool						(*ReadLong					  )(HyperFile *hf, LONG          *v);
	Bool						(*ReadULong					  )(HyperFile *hf, ULONG         *v);
	Bool						(*ReadReal					  )(HyperFile *hf, Real          *v);
	Bool						(*ReadLReal					  )(HyperFile *hf, LReal         *v);
	Bool						(*ReadBool					  )(HyperFile *hf, Bool          *v);
	Bool						(*ReadTime						)(HyperFile *hf, BaseTime      *v);
	Bool						(*ReadVector					)(HyperFile *hf, Vector        *v);
	Bool						(*ReadLVector					)(HyperFile *hf, LVector       *v);
	Bool						(*ReadMatrix					)(HyperFile *hf, Matrix        *v);
	Bool						(*ReadLMatrix					)(HyperFile *hf, LMatrix       *v);
	Bool						(*ReadString					)(HyperFile *hf, _String       *v);
	Bool						(*ReadFilename				)(HyperFile *hf, _Filename     *v);
	Bool						(*ReadImage						)(HyperFile *hf, BaseBitmap   *v);
	Bool						(*ReadContainer				)(HyperFile *hf, _BaseContainer *v, Bool flush);
	Bool						(*ReadMemory					)(HyperFile *hf, void **data, LONG *size);

	LONG						(*GetError						)(HyperFile *hf);
	void						(*SetError						)(HyperFile *hf, LONG err);
	Bool						(*ReadValueHeader			)(HyperFile *hf, UCHAR *h);
	Bool						(*SkipValue						)(HyperFile *hf, UCHAR h);
	Bool						(*WriteChunkStart			)(HyperFile *hf, LONG id, LONG level);
	Bool						(*WriteChunkEnd				)(HyperFile *hf);
	Bool						(*ReadChunkStart			)(HyperFile *hf, LONG *id, LONG *level);
	Bool						(*ReadChunkEnd				)(HyperFile *hf);
	Bool						(*SkipToEndChunk			)(HyperFile *hf);
};

struct C4D_BaseContainer
{
	_BaseContainer	*Default;

	// container
	_BaseContainer* (*Alloc								)(LONG id);
	void						(*Free								)(_BaseContainer *killme);

	_BaseContainer* (*GetClone						)(_BaseContainer *bc);
	void						(*FlushAll						)(_BaseContainer *bc);
	LONG						(*GetId								)(_BaseContainer *bc);
	void						(*SetId								)(_BaseContainer *bc, LONG newid);
	Bool						(*Compare							)(_BaseContainer *bc1,_BaseContainer *bc2);

	Bool						(*SetReal							)(_BaseContainer *bc, LONG id, Real v);
	Bool						(*SetBool							)(_BaseContainer *bc, LONG id, Bool v);
	Bool						(*SetLong							)(_BaseContainer *bc, LONG id, LONG v);
	Bool						(*SetString						)(_BaseContainer *bc, LONG id, _String *v);
	Bool						(*SetFilename					)(_BaseContainer *bc, LONG id, _Filename *v);
	Bool						(*SetBaseTime					)(_BaseContainer *bc, LONG id, const BaseTime &v);
	Bool						(*SetContainer				)(_BaseContainer *bc, LONG id, _BaseContainer *v);
	Bool						(*SetVector						)(_BaseContainer *bc, LONG id, const Vector &v);
	Bool						(*SetMatrix						)(_BaseContainer *bc, LONG id, const Matrix &v);

	Real						(*GetReal							)(_BaseContainer *bc, LONG id);
	Bool						(*GetBool							)(_BaseContainer *bc, LONG id);
	LONG						(*GetLong							)(_BaseContainer *bc, LONG id);
	_String*				(*GetString						)(_BaseContainer *bc, LONG id);
	_Filename*			(*GetFilename					)(_BaseContainer *bc, LONG id);
	BaseTime				(*GetBaseTime					)(_BaseContainer *bc, LONG id);
	_BaseContainer* (*GetContainer			  )(_BaseContainer *bc, LONG id);
	Vector					(*GetVector						)(_BaseContainer *bc, LONG id);
	Matrix					(*GetMatrix					  )(_BaseContainer *bc, LONG id);
	Bool						(*RemoveData					)(_BaseContainer *bc, LONG id);
	LONG						(*FindIndex						)(_BaseContainer *bc, LONG id);
	LONG						(*GetIndexId					)(_BaseContainer *bc, LONG index);
	Bool						(*RemoveIndex					)(_BaseContainer *bc, LONG i);
	Bool						(*CopyTo							)(_BaseContainer *bc, _BaseContainer *dest);
	LONG						(*GetIndexType				)(_BaseContainer *bc, LONG index);

	GeData*					(*GetIndexData				)(_BaseContainer *bc, LONG index);
	Bool						(*InsData							)(_BaseContainer *bc, LONG id, GeData *n);
	Bool						(*SetData							)(_BaseContainer *bc, LONG id, GeData *n);
};

struct C4D_String
{
	// safety value
	_String					*Default;

	// strings
	_String*				(*Alloc								)(void);
	_String*				(*AllocCString				)(CHAR *txt, LONG count, StringEncoding type);
	_String*				(*AllocArray					)(LONG count, UWORD fillchr);
	void						(*Free								)(_String *str);
 	_String*				(*GetClone						)(_String *src);
	LONG						(*GetLength						)(_String *str);
	Bool						(*Add									)(_String *dst,_String *src);

	LONG						(*Compare							)(_String *str1,_String *str2,LONG flag);
	Bool						(*FindFirst						)(_String *str,_String *find, LONG *pos, LONG start);
	Bool						(*FindLast						)(_String *str,_String *find, LONG *pos, LONG start);
  Bool						(*Delete							)(_String *str,LONG pos,LONG count);
	void						(*Insert							)(_String *str,LONG pos, _String *ins);
	_String*				(*SubStr							)(_String *str,LONG start, LONG count);

	Real						(*ToReal							)(_String *str,LONG *error, LONG einheit, LONG winkelart, LONG basis);
	LONG						(*ToLong							)(_String *str, LONG *error);

	_String*				(*RealToString				)(Real v, LONG vk, LONG nk, Bool e, UWORD xchar);
	_String*				(*LongToString				)(LONG l);

	_String*				(*ToUpper							)(_String *str);
	_String*				(*ToLower							)(_String *str);

	LONG						(*GetCStringLen				)(_String *str,StringEncoding type);
	LONG						(*GetCString					)(_String *str,CHAR *cstr, LONG max, StringEncoding type);

	UWORD						(*GetChr							)(_String *str,LONG pos);
	void						(*SetChr							)(_String *str,LONG pos,UWORD chr);
};

struct C4D_Bitmap
{
	BaseBitmap*			(*Alloc								)(void);
	void						(*Free								)(BaseBitmap *bm);
	BaseBitmap*			(*GetClone						)(BaseBitmap *src);
	LONG						(*GetBw								)(BaseBitmap *bm);
	LONG						(*GetBh								)(BaseBitmap *bm);
	LONG						(*GetBt								)(BaseBitmap *bm);
	LONG						(*GetBpz							)(BaseBitmap *bm);
	LONG						(*Init1								)(BaseBitmap *bm, LONG x, LONG y, LONG depth);
	LONG					  (*Init2								)(BaseBitmap *bm, _Filename *name, LONG frame, Bool *ismovie);
	void					  (*FlushAll						)(BaseBitmap *bm);
	LONG				    (*Save								)(BaseBitmap *bm, _Filename *name, LONG format, _BaseContainer *data, LONG savebits);
	void						(*GetLine							)(BaseBitmap *bm, LONG y, void *data);
	void						(*SetLine							)(BaseBitmap *bm, LONG y, void *data, LONG depth);
	void						(*GetAlphaLine				)(BaseBitmap *bm, LONG y, void *data, LONG channel);
	void						(*SetAlphaLine				)(BaseBitmap *bm, LONG y, void *data, LONG channel);
	void						(*SetCMAP							)(BaseBitmap *bm, LONG i, LONG r, LONG g, LONG b);
	void						(*ScaleIt							)(BaseBitmap *src, BaseBitmap *dst, LONG intens, Bool sample, Bool nprop);
	void						(*SetPen							)(BaseBitmap *bm, LONG r, LONG g, LONG b);
	void						(*Clear								)(BaseBitmap *bm, LONG x1, LONG y1, LONG x2, LONG y2, LONG r, LONG g, LONG b);
	void						(*Line								)(BaseBitmap *bm, LONG x1, LONG y1, LONG x2, LONG y2);
	void						(*SetPixel						)(BaseBitmap *bm, LONG x, LONG y, LONG r, LONG g, LONG b);
	void						(*GetPixel						)(BaseBitmap *bm, LONG x, LONG y, UWORD *r, UWORD *g, UWORD *b);
	LONG						(*AddChannel					)(BaseBitmap *bm, Bool internal, Bool straight);
	void						(*RemoveChannel				)(BaseBitmap *bm, LONG channel);
	void						(*GetAlphaPixel				)(BaseBitmap *bm, LONG channel, LONG x, LONG y, UWORD *val);
	void						(*SetAlphaPixel				)(BaseBitmap *bm, LONG channel, LONG x, LONG y, LONG val);
	LONG						(*SaveChannel					)(BaseBitmap *bm, LONG channel, _Filename *name);
	LONG						(*GetInternalChannel	)(BaseBitmap *bm);
	LONG						(*GetChannelCount			)(BaseBitmap *bm);
	LONG						(*GetChannelNum				)(BaseBitmap *bm, LONG num);
};

struct C4D_MovieSaver
{
	MovieSaver*			(*Alloc								)(void);
	void						(*Free								)(MovieSaver *ms);
	LONG						(*Open								)(MovieSaver *ms, _Filename *name, BaseBitmap *bm, LONG fps, LONG id, _BaseContainer *data, LONG savebits);
	LONG						(*Write								)(MovieSaver *ms, BaseBitmap *bm);
	void						(*Close								)(MovieSaver *ms);
	Bool						(*Choose							)(MovieSaver *ms, LONG id, _BaseContainer *bc);
};

struct C4D_BaseChannel
{
	// channels
	BaseChannel*		(*Alloc								)(void);
	BaseChannel*		(*GetClone						)(BaseChannel *bc);
	void						(*Free								)(BaseChannel *bc);
	Bool						(*Compare							)(BaseChannel *bc1,BaseChannel *bc2);
	LONG						(*InitTexture					)(BaseChannel *bc, InitRenderStruct *irs);
	void						(*FreeTexture					)(BaseChannel *bc);
	Vector					(*Sample							)(BaseChannel *bc, VolumeData *vd, Vector *p, Vector *delta, Vector *n, Real t, LONG tflag, Real off, Real scale);
	BaseBitmap*			(*GetBitmap						)(BaseChannel *bc);
	void						(*GetData							)(BaseChannel *bc, _BaseContainer *ct);
	void						(*SetData							)(BaseChannel *bc, _BaseContainer *ct);
	Bool						(*ReadData						)(HyperFile *hf, BaseChannel *bc);
	Bool						(*WriteData						)(HyperFile *hf, BaseChannel *bc);

	_TextureGroup*	(*DlgAlloc						)(BaseChannel *bc, CDialog *cd, CUserArea *cupreview, LONG idimage, LONG idname, LONG idbluroff, LONG idblurstr, LONG idint, LONG idinfo, LONG idpreview, LONG idedit, LONG idpopup);
	void						(*DlgFree							)(_TextureGroup *tg);
	Bool						(*DlgMessage					)(_TextureGroup *th, _BaseContainer *msg, Bool preview);

	Bool						(*Message             )(BaseChannel *bc, LONG id, void *data);
	PLUGIN_DATA*		(*GetPluginData				)(BaseChannel *bc);
	LONG						(*GetPluginID					)(BaseChannel *bc);

	CHANNELSHADER*	(*GetShader						)(BaseChannel *bc);
};

struct C4D_Filename
{
	// safety value
	_Filename				*Default;

	_Filename*			(*Alloc								)(void);
	_Filename*			(*GetClone						)(_Filename *fn);
	void						(*Free								)(_Filename *fn);
	Bool						(*FileSelect					)(_Filename *fn,LONG type, LONG flags, _String *title);
	Bool						(*Content							)(_Filename *fn);
	_String*				(*GetString						)(_Filename *fn);
	void						(*SetString						)(_Filename *fn, _String *str);
	_Filename*			(*GetDirectory				)(_Filename *fn);
	_Filename*			(*GetFile							)(_Filename *fn);
	void						(*ClearSuffix					)(_Filename *fn);
	void						(*SetSuffix						)(_Filename *fn, _String *str);
	Bool						(*CheckSuffix					)(_Filename *fn, _String *str);
	void						(*SetDirectory				)(_Filename *fn, _Filename *str);
	void						(*SetFile							)(_Filename *fn, _Filename *str);
	Bool						(*Compare							)(_Filename *fn1, _Filename *fn2);
	void						(*Add									)(_Filename *dst, _Filename *src);
};

struct C4D_BrowseFiles
{
	BrowseFiles*		(*Alloc								)(_Filename *dir, Bool calcsize);
	void						(*Free								)(BrowseFiles *bf);
	Bool						(*GetNext							)(BrowseFiles *bf);
	LONG						(*GetSize							)(BrowseFiles *bf);
	Bool						(*IsDir								)(BrowseFiles *bf);
	_Filename*			(*GetFilename					)(BrowseFiles *bf);
};

struct C4D_File
{
	BaseFile*				(*Alloc								)(void);
	void						(*Free								)(BaseFile *fl);
	Bool						(*Open								)(BaseFile *fl, _Filename *name, LONG mode, LONG error_dialog, LONG order, LONG type, LONG creator);
	Bool						(*Close								)(BaseFile *fl);
	void						(*SetOrder						)(BaseFile *fl, LONG order);
	LONG						(*ReadBytes						)(BaseFile *fl, void *data, LONG len, Bool just_try_it);
	LONG						(*WriteBytes					)(BaseFile *fl, void *data, LONG len);
	Bool						(*Seek								)(BaseFile *fl, LONG pos, LONG mode);
	LONG						(*GetPosition					)(BaseFile *fl);
	LONG						(*GetLength						)(BaseFile *fl);
	LONG						(*GetError						)(BaseFile *fl);
	void						(*SetError						)(BaseFile *fl, LONG error);
	Bool						(*WriteChar						)(BaseFile *fl, CHAR  v);
	Bool						(*WriteUChar					)(BaseFile *fl, UCHAR v);
	Bool						(*WriteWord						)(BaseFile *fl, WORD  v);
	Bool						(*WriteUWord					)(BaseFile *fl, UWORD v);
	Bool						(*WriteLong						)(BaseFile *fl, LONG  v);
	Bool						(*WriteULong					)(BaseFile *fl, ULONG v);
	Bool						(*WriteReal						)(BaseFile *fl, Real  v);
	Bool						(*WriteLReal					)(BaseFile *fl, LReal v);
	Bool						(*ReadChar						)(BaseFile *fl, CHAR  *v);
	Bool						(*ReadUChar						)(BaseFile *fl, UCHAR *v);
	Bool						(*ReadWord						)(BaseFile *fl, WORD  *v);
	Bool						(*ReadUWord						)(BaseFile *fl, UWORD *v);
	Bool						(*ReadLong						)(BaseFile *fl, LONG  *v);
	Bool						(*ReadULong						)(BaseFile *fl, ULONG *v);
	Bool						(*ReadReal						)(BaseFile *fl, Real  *v);
	Bool						(*ReadLReal						)(BaseFile *fl, LReal *v);
};

struct C4D_Dialog
{
	CDialog*				(*Alloc               )(CDialogMessage *dlgfunc,void *userdata);
	void						(*Free                )(CDialog *cd);
	void*						(*GetUserData         )(CDialog *cd);
	CUserArea*			(*AttachUserArea      )(CDialog *cd, LONG id,void *userdata);
	Bool						(*Open                )(CDialog *cd, Bool async, CDialog *parent, LONG xpos, LONG ypos);
	Bool						(*Close               )(CDialog *cd);
	Bool						(*SendUserAreaMessage )(CDialog *cd, LONG id,_BaseContainer *msg);
	Bool						(*Enable              )(CDialog *cd, LONG id, Bool enabled);
	void						(*SetTimer            )(CDialog *cd, LONG timer);
	Bool						(*SetLong             )(CDialog *cd, LONG id, LONG value,LONG min,LONG max,LONG step);
	Bool						(*SetReal             )(CDialog *cd, LONG id, Real value,Real min,Real max,Real step,LONG format);
	Bool						(*SetVector           )(CDialog *cd, LONG id, const Vector &value);
	Bool						(*SetString           )(CDialog *cd, LONG id, _String *text);
	Bool						(*SetColorField       )(CDialog *cd, LONG id, const Vector &color, Real brightness,Real maxbrightness,LONG flags);
	Bool						(*GetLong             )(CDialog *cd, LONG id, LONG &value);
	Bool						(*GetReal             )(CDialog *cd, LONG id, Real &value);
	Bool						(*GetVector           )(CDialog *cd, LONG id, Vector &value);
	Bool						(*GetString           )(CDialog *cd, LONG id, _String *&text);
	Bool						(*GetColorField       )(CDialog *cd, LONG id, Vector &color, Real &brightness);
	Bool						(*AddGadget           )(CDialog *cd, LONG type, LONG id, _String *name,LONG par1,LONG par2, LONG par3, LONG par4);
	Bool						(*LoadDialogResource  )(CDialog *cd, LONG id, LocalResource *lr, LONG flags);
	Bool						(*TabGroupBegin       )(CDialog *cd, LONG id, LONG flags,LONG tabtype);
	Bool						(*GroupBegin          )(CDialog *cd, LONG id, LONG flags,LONG cols,LONG rows,_String *title,LONG groupflags);
	Bool						(*GroupSpace          )(CDialog *cd, LONG spacex,LONG spacey);
	Bool						(*GroupBorder         )(CDialog *cd, LONG borderstyle);
	Bool						(*GroupBorderSize     )(CDialog *cd, LONG left, LONG top,LONG right,LONG bottom);
	Bool						(*GroupEnd            )(CDialog *cd);
	Bool						(*SetPopup            )(CDialog *cd, LONG id,_BaseContainer *bc);
	Bool						(*Screen2Local        )(CDialog *cd, LONG *x, LONG *y);
	Bool						(*SetVisibleArea      )(CDialog *cd, LONG scrollgroupid, LONG x1, LONG y1,LONG x2,LONG y2);
	Bool						(*GetItemDim          )(CDialog *cd, LONG id, LONG *x, LONG *y, LONG *w, LONG *h);
	Bool						(*ScrollGroupBegin    )(CDialog *cd, LONG id, LONG flags,LONG scrollflags);
	Bool						(*SendRedrawThread		)(CDialog *cd, LONG id);
	Bool            (*GetVisibleArea      )(CDialog *cd, LONG id, LONG *x1,LONG *y1,LONG *x2,LONG *y2);
	Bool						(*RestoreLayout				)(CDialog *cd, void *secret);
	Bool						(*SetMessageResult    )(CDialog *cd, _BaseContainer *result);
	Bool            (*GetDragObject       )(CDialog *cd, _BaseContainer *msg,LONG *type,void **object);

	Bool            (*SetDragDestination  )(CDialog *cd, LONG cursor);
	Bool						(*AttachSubDialog     )(CDialog *parentcd,LONG id,CDialog *cd);
};

struct C4D_UserArea
{
	void						(*Free                )(CUserArea* cu);
	void*						(*GetUserData         )(CUserArea *cu);
	LONG						(*GetWidth            )(CUserArea *cu);
	LONG						(*GetHeight           )(CUserArea *cu);
	LONG						(*GetID               )(CUserArea *cu);
	void						(*SetMinSize          )(CUserArea *cu, LONG w,LONG h);
	void						(*DrawLine            )(CUserArea *cu, LONG x1,LONG y1,LONG x2,LONG y2);
	void						(*DrawRectangle       )(CUserArea *cu, LONG x1,LONG y1,LONG x2,LONG y2);
	void						(*DrawSetPenV         )(CUserArea *cu, const Vector &color);
	void						(*DrawSetPenI         )(CUserArea *cu, LONG id);
	void						(*SetTimer            )(CUserArea *cu, LONG timer);
	Bool						(*GetInputState       )(CUserArea *cu, LONG askdevice,LONG askchannel,_BaseContainer *res);
	Bool						(*GetInputEvent       )(CUserArea *cu, LONG askdevice,_BaseContainer *res);
	void						(*KillEvents          )(CUserArea *cu);
	void						(*DrawSetFont         )(CUserArea *cu, LONG fontid);
	LONG						(*DrawGetTextWidth    )(CUserArea *cu, _String *text);
	LONG						(*DrawGetFontHeight   )(CUserArea *cu);
	void						(*DrawSetTextColII    )(CUserArea *cu, LONG fg,LONG bg);
	void						(*DrawSetTextColVI    )(CUserArea *cu, const Vector &fg,LONG bg);
	void						(*DrawSetTextColIV    )(CUserArea *cu, LONG fg,const Vector &bg);
	void						(*DrawSetTextColVV    )(CUserArea *cu, const Vector &fg,const Vector &bg);
	void						(*DrawText            )(CUserArea *cu, _String *txt,LONG x,LONG y);
	void						(*DrawBitmap          )(CUserArea *cu, BaseBitmap *bmp,LONG wx,LONG wy,LONG ww,LONG wh,LONG x,LONG y,LONG w,LONG h,LONG mode);
	void						(*SetClippingRegion   )(CUserArea *cu, LONG x,LONG y,LONG w,LONG h);
	void						(*ScrollArea          )(CUserArea *cu, LONG xdiff,LONG ydiff,LONG x,LONG y,LONG w,LONG h);
	void						(*ClearClippingRegion )(CUserArea *cu);
	Bool						(*OffScreenOn         )(CUserArea *cu);

	Bool						(*Global2Local        )(CUserArea *cu, LONG *x,LONG *y);
	Bool						(*SendParentMessage   )(CUserArea *cu, _BaseContainer *msg);

	Bool						(*Screen2Local        )(CUserArea *cu, LONG *x, LONG *y);
	Bool            (*SetDragDestination  )(CUserArea *cu, LONG cursor);
	Bool            (*HandleMouseDrag     )(CUserArea *cu, const _BaseContainer *msg,LONG type,void *data,Bool privateframe);
	Bool						(*IsEnabled           )(CUserArea *cu);

	void						(*GetBorderSize       )(CUserArea *cu,LONG type,LONG *l,LONG *t,LONG *r,LONG *b);
	void            (*DrawBorder          )(CUserArea *cu,LONG type,LONG x1,LONG y1,LONG x2,LONG y2);

	_GeListView*		(*GeListView_Alloc            )(void);
	void						(*GeListView_Free             )(_GeListView *lv);
	Bool            (*GeListView_Attach           )(_GeListView *lv,CDialog *cd,LONG id,ListViewCallBack *callback,void *userdata);
	void            (*GeListView_LvSuperCall      )(_GeListView *lv,LONG &res_type,void *&result,void *secret,LONG cmd,LONG line,LONG col);
	void            (*GeListView_Redraw           )(_GeListView *lv);
	void            (*GeListView_DataChanged      )(_GeListView *lv);
	Bool            (*GeListView_ExtractMouseInfo )(_GeListView *lv,void *secret,MouseDownInfo &info,LONG size);
	Bool            (*GeListView_ExtractDrawInfo  )(_GeListView *lv,void *secret,DrawInfo &info,LONG size);
	Bool            (*GeListView_SendParentMessage)(_GeListView *lv,_BaseContainer *msg);
	LONG						(*GeListView_GetId            )(_GeListView *lv);

	_SimpleListView* (*SimpleListView_Alloc       )(void);
	void            (*SimpleListView_Free         )(_SimpleListView *lv);
	Bool            (*SimpleListView_SetLayout    )(_SimpleListView *lv,LONG columns,const _BaseContainer &data);
	Bool            (*SimpleListView_SetItem      )(_SimpleListView *lv,LONG id,const _BaseContainer &data);
	Bool            (*SimpleListView_GetItem      )(_SimpleListView *lv,LONG id,_BaseContainer *data);
	LONG            (*SimpleListView_GetItemCount )(_SimpleListView *lv);
	Bool            (*SimpleListView_GetItemLine  )(_SimpleListView *lv,LONG num,LONG *id,_BaseContainer *data);
	Bool            (*SimpleListView_RemoveItem   )(_SimpleListView *lv,LONG id);
	LONG            (*SimpleListView_GetSelection )(_SimpleListView *lv,BaseSelect *selection);
	Bool            (*SimpleListView_SetSelection )(_SimpleListView *lv,BaseSelect *selection);

	LONG            (*SimpleListView_GetProperty  )(_SimpleListView *lv,LONG id);
	Bool            (*SimpleListView_SetProperty  )(_SimpleListView *lv,LONG id,LONG val);
};

struct C4D_Parser
{
	Parser*					(*Alloc								)(void);
	void						(*Free								)(Parser *pr);
	Bool						(*AddVar							)(Parser *pr, _String *str, Real *value);
	Bool						(*Eval								)(Parser *pr, _String *str, LONG *error,Real *res,LONG unit,LONG angletype,LONG basis);
};

struct C4D_Resource
{
	LocalResource*	(*Alloc								)(_Filename *path);
	void						(*Free                )(LocalResource *lr);
	_String*				(*LoadString          )(LocalResource *lr,LONG id);
	LocalResource*  (*GetCinemaResource   )(void);
};

struct C4D_BaseList
{
	LONG						(*GetType							)(BaseList2D *bl);
	LONG						(*GetDiskType					)(BaseList2D *bl);
	void						(*GetMarker						)(BaseList2D *bl, ULONG *l1, ULONG *l2);
	BaseList2D*			(*Next								)(BaseList2D *bl);
	BaseList2D*			(*Pred								)(BaseList2D *bl);
	void						(*InsertBefore				)(BaseList2D *element, BaseList2D *bl);
	void						(*InsertAfter					)(BaseList2D *element, BaseList2D *bl);
	void						(*Remove							)(BaseList2D *bl);
	void						(*SetAllBits					)(BaseList2D *bl, LONG mask);
	LONG						(*GetAllBits					)(BaseList2D *bl);
	void						(*SetColor						)(BaseList2D *bl, LONG mask);
	LONG						(*GetColor						)(BaseList2D *bl);
	Bool						(*Message							)(BaseList2D *bl, LONG type, void *data);
	BaseList4D*			(*Up									)(BaseList4D *bl);
	BaseList4D*			(*Down								)(BaseList4D *bl);
	BaseList4D*			(*DownLast						)(BaseList4D *bl);
	void						(*InsertUnder					)(BaseList4D *element, BaseList4D *bl);
	void						(*InsertUnderLast			)(BaseList4D *element, BaseList4D *bl);
	void						(*Free								)(BaseList2D *bl);
	void*						(*GetMain							)(BaseList2D *bl);
};

struct C4D_Tag
{
	BaseTag*				(*Alloc								)(LONG type, LONG subtype, LONG count);
	BaseTag*				(*GetClone						)(BaseTag *bt, LONG flags);
	Bool						(*CopyTo							)(BaseTag *bt, BaseTag *dest, LONG flags);
	LONG						(*GetTagInfo					)(BaseTag *bt);
	Bool						(*Edit								)(BaseTag *bt);
	LONG						(*GetDataCount				)(VariableTag *bt);
	LONG						(*GetDataSize					)(VariableTag *bt);
	void*						(*GetDataAddress			)(VariableTag *bt);
	void						(*SetData							)(BaseTag *tag, _BaseContainer *bc);
	void						(*GetData							)(BaseTag *tag, _BaseContainer *bc);
	BaseSelect*			(*GetBaseSelect				)(SelectionTag *tag);
	Bool						(*Record							)(StickTextureTag *stt, BaseObject *op, Bool always);
	LONG						(*GetPluginID					)(CPluginTag *tag);
	void*						(*GetPluginData				)(CPluginTag *tag);
	void						(*SetPluginData				)(CPluginTag *tag,void *mem);

	// UVWs
	void						(*UvGet								)(UVWTag *tag, LONG i, UVWStruct *s);
	void						(*UvSet								)(UVWTag *tag, LONG i, UVWStruct *s);
	void						(*UvCpy								)(UVWTag *tag, LONG dst, UVWTag *srctag, LONG src);

	void						(*GetIcon							)(BaseTag *tag, IconData *dat);
};

struct C4D_Object
{
	BaseObject*			(*Alloc								)(LONG type, LONG subtype);
	SplineObject*		(*AllocSplineObject		)(LONG pcnt, LONG type);
	PluginObject*		(*AllocPluginObject		)(LONG plug_id);
	LONG						(*GetSubType					)(BaseObject *op);
	LONG						(*GetLayerType				)(BaseObject *op);
	_BaseContainer*	(*GetData							)(BaseObject *op);
	void						(*GetParameter				)(BaseObject *op, _BaseContainer *data);
	void						(*SetParameter				)(BaseObject *op, _BaseContainer *data);
	Real						(*GetVisibility				)(BaseObject *op, Real parent);
	void						(*SetVisibility				)(BaseObject *op, Real p);
	Vector					(*GetPos							)(BaseObject *op);
	void						(*SetPos							)(BaseObject *op, const Vector &v);
	Vector					(*GetScale						)(BaseObject *op);
	void						(*SetScale						)(BaseObject *op, const Vector &v);
	Vector					(*GetRot							)(BaseObject *op);
	void						(*SetRot							)(BaseObject *op, const Vector &v);
	Matrix					(*GetMl								)(BaseObject *op);
	void						(*SetMl								)(BaseObject *op, const Matrix &m);
	Matrix					(*GetMg								)(BaseObject *op);
	void						(*SetMg								)(BaseObject *op, const Matrix &m);
	Matrix					(*GetMln							)(BaseObject *op);
	Matrix					(*GetMgn							)(BaseObject *op);
	Matrix					(*GetUpMg							)(BaseObject *op);
	Vector					(*GetMp								)(BaseObject *op);
	Vector					(*GetRad							)(BaseObject *op);
	LONG						(*GetMode							)(BaseObject *op, LONG mode);
	void						(*SetMode							)(BaseObject *op, LONG mode, LONG val);
	void						(*GetName							)(BaseObject *op, _String *name);
	void						(*SetName							)(BaseObject *op, _String *name);
	BaseTag*				(*GetFirstTag					)(BaseObject *op);
	BaseTag*				(*GetTag							)(BaseObject *op, LONG type, LONG nr);
	void*						(*GetTagData					)(BaseObject *op, LONG type, LONG nr);
	LONG						(*GetTagDataCount			)(BaseObject *op, LONG type);
	void						(*InsertTag						)(BaseObject *op, BaseTag *tp, BaseTag *pred);
	void						(*KillTag							)(BaseObject *op, LONG type, LONG nr);
	BaseTrack*			(*GetFirstTrack				)(BaseObject *op);
	BaseTrack*			(*GetTrack						)(BaseObject *op, LONG type);
	void						(*KillAllTracks				)(BaseObject *op);
	void						(*InsertTrack					)(BaseObject *op, BaseTrack *track, BaseTrack *pred);
	LONG						(*GetObjectInfo				)(BaseObject *op);
	Bool						(*Edit								)(BaseObject *op);
	BaseObject*			(*GetCache						)(BaseObject *op, HierarchyHelp *hh);
	BaseObject*			(*GetDeformCache			)(BaseObject *op);
	LineObject*			(*GetIsoparm					)(BaseObject *op);
	Bool						(*CopyTo							)(BaseObject *op, BaseObject *dest, LONG flags);
	Bool						(*CopyTagsTo					)(BaseObject *op, BaseObject *dest, LONG visible, LONG variable, LONG hierarchical);
	BaseObject*			(*GetClone						)(BaseObject *op, LONG flags);
	Bool						(*IsDirty							)(BaseObject *op, LONG flags);
	void						(*SetDirty						)(BaseObject *op, LONG flags);
	Bool						(*CheckCache					)(BaseObject *op, HierarchyHelp *hh);
	void						(*SetIsoparm					)(BaseObject *op, LineObject *l);
	BaseObject*			(*GenPrimitive				)(BaseDocument *doc, LONG type, _BaseContainer *bc, Real lod, Bool isoparm, BaseThread *bt);
	BaseObject*			(*GenSplinePrimitive	)(BaseDocument *doc, LONG type, _BaseContainer *bc, Real lod, BaseThread *bt);
	LONG						(*GetPluginID					)(PluginObject *op);
	void*						(*GetPluginData				)(PluginObject *op);
	void						(*SetPluginData				)(PluginObject *op,void *mem);
	Bool						(*IsAddictedEx				)(BaseObject *op,BaseDocument *doc);
	Bool						(*GetActiveDependence )(BaseObject *op);
	void						(*SetActiveDependence )(BaseObject *op, Bool on);
	void						(*NewDependenceList		)(PluginObject *op);
	Bool						(*CmpDependenceList		)(PluginObject *op);
	void						(*TouchDependenceList	)(PluginObject *op);
	void						(*AddDependence				)(PluginObject *op,HierarchyHelp *hh, BaseObject *pp);
	Bool						(*AddTexture					)(PluginObject *op,_Filename *fn, void *priv);
	BaseObject*			(*GetHierarchyClone		)(PluginObject *op,HierarchyHelp *hh, BaseObject *pp, LONG flags, Bool *dirty);
	BaseObject*			(*GetVirtualLineObject)(BaseObject *op, HierarchyHelp *hh, const Matrix &mloc, Bool keep_spline, Bool recurse, Matrix *mres, Bool *dirty);
	void						(*Touch								)(BaseObject *op);

	// point object
	BaseSelect*			(*PoGetPointS					)(PointObject *op);
	BaseSelect*			(*PoGetPointH					)(PointObject *op);
	Bool						(*PoResizeObject			)(PointObject *op, LONG pcnt);
	Real*						(*PoCalcVertexMap			)(PointObject *op, BaseObject *modifier);

	// line object
	Bool						(*LoResizeObject			)(LineObject *op, LONG pcnt, LONG scnt);

	// polygon object
	BaseSelect*			(*PyGetPolygonS				)(PolygonObject *op);
	BaseSelect*			(*PyGetPolygonH				)(PolygonObject *op);
	Bool						(*PyResizeObject			)(PolygonObject *op, LONG pcnt, LONG vcnt);

	// spline object
	Vector					(*SpGetSplinePoint		)(SplineObject *op, Real t, LONG segment, Vector *padr);
	Vector					(*SpGetSplineTangent	)(SplineObject *op, Real t, LONG segment, Vector *padr);
	Bool						(*SpInitLength				)(SplineObject *op, LONG segment, Vector *padr);
	void						(*SpFreeLength				)(SplineObject *op);
	Real						(*SpUniformToNatural	)(SplineObject *op, Real t);
	Real						(*SpGetLength					)(SplineObject *op);
	Real						(*SpGetSegmentLength	)(SplineObject *op, LONG a, LONG b);
	LineObject*			(*SpGetLineObject			)(SplineObject *op, BaseDocument *doc, Real lod, BaseThread *thread);
	SplineObject*		(*SpGetRealSpline			)(SplineObject *op);
	Bool						(*SpResizeObject			)(SplineObject *op, LONG pcnt, LONG scnt);

	// particle object
	LONG						(*PrGetCount					)(ParticleObject *op);
	Real						(*PrGetLifetime				)(ParticleObject *op);
	Particle*				(*PrGetParticle				)(ParticleObject *op, ParticleTag *pt, LONG i);
	Bool						(*PrIsMatrixAvailable	)(ParticleObject *op);
	ParticleDetails*(*PrGetParticleDetails)(BaseDocument *doc, ParticleObject *op);

	// camera object
	LONG						(*CoGetProjection			)(BaseObject *op);
	Real						(*CoGetFocus					)(BaseObject *op);
	Real						(*CoGetZoom						)(BaseObject *op);
	Vector					(*CoGetOffset					)(BaseObject *op);
	Real						(*CoGetAperture				)(BaseObject *op);
	Bool						(*CoSetProjection			)(BaseObject *op, LONG projection);
	Bool						(*CoSetFocus					)(BaseObject *op, Real v);
	Bool						(*CoSetZoom						)(BaseObject *op, Real v);
	Bool						(*CoSetOffset					)(BaseObject *op, const Vector &offset);
	Bool						(*CoSetAperture				)(BaseObject *op, Real v);

	// object safety
	ObjectSafety*		(*OsAlloc							)(BaseObject *op);
	void						(*OsFree							)(ObjectSafety *os, Bool restore);

	// triangulation
	Bool						(*Triangulate					)(Vector *padr, LONG pcnt, Polygon **vadr, LONG *vcnt);
	PolygonObject*	(*TriangulateLine			)(LineObject *op, Real regular, BaseThread *bt);
	SplineObject*		(*FitCurve						)(Vector *padr, LONG pcnt, Real error, BaseThread *bt);

	// uv stuff
	UVWTag*					(*GenerateUVW					)(BaseObject *op, const Matrix &opmg, TextureTag *tp, const Matrix &texopmg, BaseView *view);

	void						(*GetIcon							)(BaseObject *op, IconData *dat);

	LONG						(*GetDirty						)(BaseObject *op, LONG flags);
};

struct C4D_Document
{
	// render data
	RenderData*			(*RdAlloc							)(void);
	void						(*RdGetData						)(RenderData *rd, _BaseContainer *bc);
	void						(*RdSetData						)(RenderData *rd, _BaseContainer *bc);

	// document
	BaseDocument*		(*Alloc								)(void);
	Bool						(*CopyTo							)(BaseDocument *src, BaseDocument *dst, LONG flags);
	void						(*FlushAll						)(BaseDocument *doc);
	void						(*GetData							)(BaseDocument *doc, LONG type, _BaseContainer *bc);
	void						(*SetData							)(BaseDocument *doc, LONG type, _BaseContainer *bc);
	BaseObject*			(*GetFirstObject			)(BaseDocument *doc);
	BaseMaterial*		(*GetFirstMaterial		)(BaseDocument *doc);
	RenderData*			(*GetFirstRenderData	)(BaseDocument *doc);
	void						(*InsertRenderData		)(BaseDocument *doc, RenderData *rd, RenderData *pred);
	void						(*InsertMaterial			)(BaseDocument *doc, BaseMaterial *mat, BaseMaterial *pred, Bool checknames);
	void						(*InsertObject				)(BaseDocument *doc, BaseObject *op, BaseObject *parent, BaseObject *pred, Bool checknames);
	RenderData*			(*GetActiveRenderData	)(BaseDocument *doc);
	BaseObject*			(*GetActiveObject			)(BaseDocument *doc);
	BaseMaterial*		(*GetActiveMaterial		)(BaseDocument *doc);
	BaseTag*				(*GetActiveTag				)(BaseDocument *doc, BaseObject *active);
	void						(*SetActiveObject			)(BaseDocument *doc, BaseObject *op);
	void						(*SetActiveTag				)(BaseDocument *doc, BaseObject *op, BaseTag *ta);
	void						(*SetActiveMaterial		)(BaseDocument *doc, BaseMaterial	*mp);
	void						(*SetActiveRenderData	)(BaseDocument *doc, RenderData *rd);
	BaseObject*			(*GetHighest					)(BaseDocument *doc, LONG type, Bool editor);
	BaseMaterial*		(*SearchMaterial			)(BaseDocument *doc, _String *str, Bool inc);
	BaseObject*			(*SearchObject				)(BaseDocument *doc, _String *str, Bool inc);
	Bool						(*StartUndo						)(BaseDocument *doc);
	Bool						(*EndUndo							)(BaseDocument *doc);
	Bool						(*AddUndo							)(BaseDocument *doc, LONG type, void *data);
	Bool						(*DoUndo							)(BaseDocument *doc);
	Bool						(*DoRedo							)(BaseDocument *doc);
	void						(*AnimateDocumentEx		)(BaseDocument *doc, LONG flags);
	void						(*AnimateObject				)(BaseDocument *doc, BaseObject *op, const BaseTime &time, LONG flags);
	BaseDraw*				(*GetActiveBaseDraw		)(BaseDocument *doc);
	BaseDraw*				(*GetRenderBaseDraw		)(BaseDocument *doc);
	BaseDraw*				(*GetBaseDraw					)(BaseDocument *doc, LONG num);
	LONG						(*GetSplinePlane			)(BaseDocument *doc);

	// hierarchy help
	Real						(*HhGetLOD						)(HierarchyHelp *hh);
	LONG						(*HhGetVFlags					)(HierarchyHelp *hh);
	BaseThread*			(*HhGetThread					)(HierarchyHelp *hh);
	BaseDocument*		(*HhGetDocument				)(HierarchyHelp *hh);
	Matrix					(*HhGetMg							)(HierarchyHelp *hh);
	void						(*HhAddVFlags					)(HierarchyHelp *hh, LONG mask);

	// hierarchy
	Bool						(*RunHierarchy				)(void *main, BaseDocument *doc, Bool spheres, Real lod, LONG flags, void *startdata, BaseThread *bt, HierarchyAlloc *ha, HierarchyFree *hf, HierarchyCopyTo *hc, HierarchyDo *hd);

	Bool						(*AnimateDocument     )(BaseDocument *doc,BaseThread *bt,Bool doexpressions,Bool doanim);
};

struct C4D_Thread
{
	BaseThread*			(*Alloc								)(ThreadMain *tm, ThreadTest *tt, void *data);
	void						(*Free								)(BaseThread *bt);
	Bool						(*TestBreak						)(BaseThread *bt);
	Bool						(*TestBaseBreak				)(BaseThread *bt);
	Bool						(*IsRunning						)(BaseThread *bt);
	Bool						(*Start								)(BaseThread *bt, Bool back, Bool pri);
	void						(*End									)(BaseThread *bt);
	void						(*ThreadLock					)(void);
	void						(*ThreadUnLock				)(void);
	LONG						(*GetCPUCount					)(void);
	MPBaseThread*		(*MPAlloc							)(BaseThread *parent, LONG count, ThreadMain *tm, ThreadTest *tt, void **data);
	void						(*MPFree							)(MPBaseThread *mp);
	BaseThread*			(*MPGetThread					)(MPBaseThread *mp, LONG i);
	void*						(*MPWaitForNextFree		)(MPBaseThread *mp);
	void						(*MPWait							)(MPBaseThread *mp);
	void						(*MPEnd								)(MPBaseThread *mp);
	Semaphore*			(*SMAlloc							)(void);
	void						(*SMFree							)(Semaphore *sm);
	Bool						(*SMLock							)(Semaphore *sm);
	Bool						(*SMLockAndWait				)(Semaphore *sm, BaseThread *bt);
	void						(*SMUnLock						)(Semaphore *sm);
	void						(*Wait								)(BaseThread *bt, Bool checkevents);
};

struct C4D_Material
{
	BaseMaterial*		(*Alloc								)(LONG type, LONG shader_id);
	LONG						(*GetPluginID					)(PluginMaterial *mat);
	PLUGIN_DATA*		(*GetPluginData				)(PluginMaterial *mat);
	void						(*GetName							)(BaseMaterial *mat, _String *name);
	void						(*SetName							)(BaseMaterial *mat, _String *name);

	void						(*Update							)(BaseMaterial *mat, LONG preview, Bool rttm);
	BaseMaterial*		(*GetClone						)(BaseMaterial *bm, LONG flags);
	Bool						(*CopyTo							)(BaseMaterial *bm, BaseMaterial *dest, LONG flags);
	Bool						(*Edit								)(BaseMaterial *bm);
	void						(*SetData							)(BaseMaterial *bm, _BaseContainer *bc);
	void						(*GetData							)(BaseMaterial *bm, _BaseContainer *bc);
	BaseChannel*		(*GetChannel					)(BaseMaterial *bm, LONG id);
	Bool						(*GetChannelState			)(Material *mat, LONG channel);
	void						(*SetChannelState			)(Material *mat, LONG channel, Bool state);
	Bool						(*Compare							)(BaseMaterial *m1, BaseMaterial *m2);

	VOLUMESHADER*		(*GetShader						)(PluginMaterial *mat);
};

struct C4D_Texture
{
	Vector					(*GetPos							)(BaseList2D *tag);
	Vector					(*GetScale						)(BaseList2D *tag);
	Vector					(*GetRot							)(BaseList2D *tag);
	Matrix					(*GetMl								)(BaseList2D *tag);
	Matrix					(*GetMln							)(BaseList2D *tag);
	void						(*SetPos							)(BaseList2D *tag, const Vector &v);
	void						(*SetScale						)(BaseList2D *tag, const Vector &v);
	void						(*SetRot							)(BaseList2D *tag, const Vector &v);
	void						(*SetMl								)(BaseList2D *tag, const Matrix &m);

	BaseMaterial*		(*GetMaterial					)(BaseList2D *tag);
	void						(*SetMaterial					)(BaseList2D *tag, BaseMaterial *mat);

	BaseMaterial*		(*GetSecondMaterial		)(TextureTag *tag, Real *fade);
	void						(*SetSecondMaterial		)(TextureTag *tag, BaseMaterial *mat, Real fade);

	BaseMaterial*   (*GetInterpolatedMaterial)(BaseDocument *doc, TextureTag *tag);
};

struct C4D_BaseSelect
{
	BaseSelect*			(*Alloc								)(void);
	void						(*Free								)(BaseSelect *bs);
	void						(*FlushAll						)(BaseSelect *bs);
	LONG						(*GetCount						)(BaseSelect *bs);
	LONG						(*GetSegments					)(BaseSelect *bs);
	Bool						(*Select							)(BaseSelect *bs, LONG num);
	Bool						(*SelectAll						)(BaseSelect *bs, LONG min, LONG max);
	Bool						(*Deselect						)(BaseSelect *bs, LONG num);
	Bool						(*DeselectAll					)(BaseSelect *bs);
	Bool						(*Toggle							)(BaseSelect *bs, LONG num);
	Bool						(*ToggleAll						)(BaseSelect *bs, LONG min, LONG max);
	Bool						(*GetRange						)(BaseSelect *bs, LONG seg, LONG *a, LONG *b);
	Bool						(*IsSelected					)(BaseSelect *bs, LONG num);
	Bool						(*CopyTo							)(BaseSelect *bs, BaseSelect *dest);
	BaseSelect*			(*GetClone						)(BaseSelect *bs);
	Bool						(*FromArray						)(BaseSelect *bs, UCHAR *selection, LONG count);
	UCHAR*					(*ToArray							)(BaseSelect *bs, LONG count);
};

struct C4D_Key
{
	BaseKey*				(*Alloc								)(LONG type, LONG plugid);
	BaseKey*				(*GetClone						)(BaseKey *bk, LONG flags);
	Bool						(*CopyTo							)(BaseKey *bk, BaseKey *dest, LONG flags);
	Bool						(*Edit								)(BaseKey *bk);
	BaseTime				(*GetTime							)(BaseKey *bk);
	void						(*SetTime							)(BaseKey *bk, const BaseTime &time);
	void						(*GetData							)(BaseKey *bk, _BaseContainer *ct);
	void						(*SetData							)(BaseKey *bk, _BaseContainer *ct);
	void						(*SetP						    )(PSRKey *bk, const Vector &v);
	void						(*SetVL						    )(PSRKey *bk, const Vector &v);
	void						(*SetVR						    )(PSRKey *bk, const Vector &v);
	Vector					(*GetP						    )(PSRKey *bk);
	Vector					(*GetVL						    )(PSRKey *bk);
	Vector					(*GetVR						    )(PSRKey *bk);
	void						(*SoftTangents				)(PSRKey *bk, BaseDocument *doc, LONG component);
	VariableTag*		(*GetPointTag					)(PLAKey *bk);
	VariableTag*		(*GetHermiteTag				)(PLAKey *bk);
	LONG						(*GetPluginID					)(CPluginKey *op);
	void*						(*GetPluginData				)(CPluginKey *op);
	void						(*SetPluginData				)(CPluginKey *op,void *mem);
	Bool						(*FillKey             )(BaseKey *bk);
};

struct C4D_Sequence
{
	BaseSequence*		(*Alloc								)(LONG type, LONG plugid);
	BaseSequence*		(*GetClone						)(BaseSequence *bs, LONG flags);
	Bool						(*CopyTo							)(BaseSequence *bs, BaseSequence *dest, LONG flags);
	Bool						(*Edit								)(BaseSequence *bs);
	BaseTime				(*GetT1								)(BaseSequence *bs);
	void						(*SetT1								)(BaseSequence *bs, const BaseTime &time);
	BaseTime				(*GetT2								)(BaseSequence *bs);
	void						(*SetT2								)(BaseSequence *bs, const BaseTime &time);
	BaseTime				(*GetT3								)(BaseSequence *bs);
	void						(*SetT3								)(BaseSequence *bs, const BaseTime &time);
	Bool						(*GetSoft							)(BaseSequence *bs);
	void						(*SetSoft							)(BaseSequence *bs, Bool b);
	void						(*FlushAllKeys				)(BaseSequence *bs);
	void						(*FlushAllTimeKeys		)(BaseSequence *bs);
	LONG						(*GetKeyCount					)(BaseSequence *bs);
	LONG						(*GetTimeKeyCount			)(BaseSequence *bs);
	BaseKey*				(*FindKey							)(BaseSequence *bs, const BaseTime &time);
	BaseTime				(*GetLeft							)(BaseSequence *bs);
	BaseTime				(*GetRight						)(BaseSequence *bs);
	LONG						(*GetTimeMode					)(BaseSequence *bs);
	void						(*SetTimeMode					)(BaseSequence *bs, LONG l);
	Bool						(*GetLeftInfluence		)(BaseSequence *bs);
	void						(*SetLeftInfluence		)(BaseSequence *bs, Bool b);
	Bool						(*GetRightInfluence		)(BaseSequence *bs);
	void						(*SetRightInfluence		)(BaseSequence *bs, Bool b);
	Real						(*GetV0								)(BaseSequence *bs);
	void						(*SetV0								)(BaseSequence *bs, Real v0);
	Real						(*GetS0								)(BaseSequence *bs);
	void						(*SetS0								)(BaseSequence *bs, Real s0);
	BaseKey*				(*GetFirstKey					)(BaseSequence *bs);
	BaseKey*				(*GetFirstTimeKey			)(BaseSequence *bs);
	Bool						(*InsertKey						)(BaseSequence *bs, BaseKey *bk);
	Bool						(*InsertTimeKey				)(BaseSequence *bs, BaseKey *bk);
	void						(*GetName							)(FolderSequence *bk, _String *ct);
	void						(*SetName							)(FolderSequence *bk, _String *ct);
	BaseObject*			(*GetFirstObject			)(FolderSequence *bk);
	void						(*InsertObject				)(FolderSequence *bk, BaseObject *op, BaseObject *parent, BaseObject *prev);
	BaseSound*			(*GetSound						)(SoundSequence *ss);
	void						(*SetSound						)(SoundSequence *ss, BaseSound *bs);
	LONG						(*GetPluginID					)(CPluginSequence *op);
	void*						(*GetPluginData				)(CPluginSequence *op);
	void						(*SetPluginData				)(CPluginSequence *op,void *mem);
	void						(*GetData							)(BaseSequence *bk, _BaseContainer *ct);
	void						(*SetData							)(BaseSequence *bk, _BaseContainer *ct);
};

struct C4D_Track
{
	BaseTrack*			(*Alloc								)(LONG type, LONG plugid);
	BaseTrack*			(*GetClone						)(BaseTrack *bt, LONG flags);
	Bool						(*CopyTo							)(BaseTrack *bt, BaseTrack *dest, LONG flags);
	void						(*GetName							)(BaseTrack *bt, _String *st);
	BaseSequence*		(*GetFirstSequence		)(BaseTrack *bt);
	Bool						(*InsertSequence			)(BaseTrack *bt, BaseSequence *bs, Bool check);
	Bool						(*AnimateObject				)(BaseTrack *bt, BaseDocument *doc, BaseObject *op, const BaseTime &time, LONG flags, void *data);
	BaseSequence*		(*FindSequence				)(BaseTrack *bt, const BaseTime &time);
	LONG						(*GetPluginID					)(CPluginTrack *op);
	void*						(*GetPluginData				)(CPluginTrack *op);
	void						(*SetPluginData				)(CPluginTrack *op,void *mem);
	void						(*GetData							)(BaseTrack *bk, _BaseContainer *ct);
	void						(*SetData							)(BaseTrack *bk, _BaseContainer *ct);
};

struct C4D_BaseSound
{
	BaseSound*			(*Alloc								)(void);
	void						(*Free								)(BaseSound *bs);
	BaseSound*			(*GetClone						)(BaseSound *bs);
	Bool						(*CopyTo							)(BaseSound *bs, BaseSound *dest);
	Bool						(*Init								)(BaseSound *bs, LONG samples, LONG fmode, Bool stereo, Bool b16);
	void						(*FlushAll						)(BaseSound *bs);
	Bool						(*Load								)(BaseSound *bs, _Filename *fn);
	Bool						(*Save								)(BaseSound *bs, _Filename *fn);
	void						(*GetSoundInfo				)(BaseSound *bs, Bool *stereo, Bool *b16, LONG *frequency, LONG *samples, BaseTime *length);
	void						(*GetSample						)(BaseSound *bs, LONG i, SData *data);
	void						(*SetSample						)(BaseSound *bs, LONG i, SData *data);
	BaseBitmap*			(*GetBitmap						)(BaseSound *bs, LONG width, LONG height, const BaseTime &start, const BaseTime &stop);
	MarkerKey*			(*GetFirstMarker			)(BaseSound *bs);
	Bool						(*InsertMarker				)(BaseSound *bs, MarkerKey *mk, MarkerKey *pred=NULL);
};

struct C4D_BaseDraw
{
	// basedraw
	Bool						(*HasCameraLink				)(BaseDraw *bd);
	void						(*SetSceneCamera			)(BaseDraw *bd, BaseObject *op);
	BaseObject*			(*GetSceneCameraEx		)(BaseDraw *bd);
	BaseObject*			(*GetEditorCamera			)(BaseDraw *bd);
	Vector					(*GetObjectColor			)(BaseDraw *bd, BaseObject *op);
	Vector					(*CheckColor					)(BaseDraw *bd, const Vector &col);
	Bool						(*ZBufAvailable				)(BaseDraw *bd);
	BaseBitmap*			(*GetBitmapEx					)(BaseDraw *bd);
	void						(*SetTransparency			)(BaseDraw *bd, LONG trans);
	LONG						(*GetTransparency			)(BaseDraw *bd);
	Bool						(*PointInRange				)(BaseDraw *bd, const Vector &p, LONG x, LONG y);
	void						(*SetPen							)(BaseDraw *bd, const Vector &col);
	Real						(*SimpleShade					)(BaseDraw *bd, const Vector &p, const Vector &n);
	void						(*Begin2D							)(BaseDraw *bd);
	void						(*End2D								)(BaseDraw *bd);
	void						(*Point2D							)(BaseDraw *bd, const Vector &p);
	void						(*Line2D							)(BaseDraw *bd, const Vector &p1, const Vector &p2); // draw line with 2D clipping
	void						(*Handle2D						)(BaseDraw *bd, const Vector &p, LONG type);
	void						(*Circle2D						)(BaseDraw *bd, LONG mx, LONG my, Real rad);
	void						(*LineCSEx						)(BaseDraw *bd, const Vector &p1, const Vector &p2, Bool clip2d, Bool clipz);
	void						(*Point3D							)(BaseDraw *bd, const Vector &p);
	void						(*Line3D							)(BaseDraw *bd, const Vector &p1, const Vector &p2);
	void						(*Handle3D						)(BaseDraw *bd, const Vector &p, LONG type);
	void						(*Circle3D						)(BaseDraw *bd, const Matrix &m);
	void						(*Box3D								)(BaseDraw *bd, const Matrix &m, Real size, const Vector &col);
	void						(*Polygon3D						)(BaseDraw *bd, Vector *p, Vector *f, Bool quad);
	void						(*PolygonObject				)(BaseDraw *bd, BaseDrawHelp *bh, PolygonObject *op, Bool force_lines);
	void						(*GetData							)(BaseDraw *bd, _BaseContainer *bc);

	// basedraw help
	BaseDocument*		(*BbGetDocument				)(BaseDrawHelp *bb);
	BaseTag*				(*BbGetActiveTag			)(BaseDrawHelp *bb);
	Matrix					(*BbGetMg							)(BaseDrawHelp *bb);
	void						(*BbGetDisplay				)(BaseDrawHelp *bb, _BaseContainer *bc);

	BaseObject*			(*GetSceneCamera			)(BaseDraw *bd, BaseDocument *doc);
	void						(*LineCS							)(BaseDraw *bd, const Vector &p1, const Vector &p2, Bool clip2d, Bool clipz,Bool offset);
};

struct C4D_BaseView
{
	void						(*GetFrame						)(BaseView *bv, LONG *cl, LONG *ct, LONG *cr, LONG *cb);
	void						(*GetSafeFrame				)(BaseView *bv, LONG *from, LONG *to, LONG *horizontal);
	void						(*GetParameter				)(BaseView *bv, Vector *offset, Vector *scale, Vector *scale_z);
	Matrix					(*GetMg								)(BaseView *bv);
	Matrix					(*GetMi								)(BaseView *bv);
	LONG						(*GetProjection				)(BaseView *bv);
	Bool						(*TestPoint						)(BaseView *bv, Real x, Real y);
	Bool						(*TestPointZ					)(BaseView *bv, const Vector &p);
	Bool						(*TestClipping3D			)(BaseView *bv, const Vector &mp, const Vector &rad, const Matrix &mg, Bool *clip2d, Bool *clipz);
	Bool						(*ClipLine2D					)(BaseView *bv, Vector *p1, Vector *p2);
	Bool						(*ClipLineZ						)(BaseView *bv, Vector *p1, Vector *p2);
	Vector					(*WS									)(BaseView *bv, const Vector &p);
	Vector					(*SW									)(BaseView *bv, const Vector &p);
	Vector					(*SW_R								)(BaseView *bv, Real x, Real y, const Vector &wp);
	Vector					(*WC									)(BaseView *bv, const Vector &p);
	Vector					(*CW									)(BaseView *bv, const Vector &p);
	Vector					(*SC									)(BaseView *bv, const Vector &p);
	Vector					(*CS									)(BaseView *bv, const Vector &p, Bool z_inverse);
	Vector					(*WC_V								)(BaseView *bv, const Vector &v);
	Vector					(*CW_V								)(BaseView *bv, const Vector &v);
	Bool						(*BackfaceCulling			)(BaseView *bv, const Vector &n, const Vector &p);
	Bool						(*ZSensitive					)(BaseView *bv);
};

struct C4D_Pool
{
	MemoryPool*			(*Alloc								)(LONG block_size);
	void						(*Free								)(MemoryPool *pool);
	void*						(*AllocElement				)(MemoryPool *pool, LONG size, Bool clear);
	void						(*FreeElement					)(MemoryPool *pool, void *mem, LONG size);
	void*						(*AllocElementS				)(MemoryPool *pool, LONG size, Bool clear);
	void						(*FreeElementS				)(MemoryPool *pool, void *mem);
};

struct C4D_General
{
	void*						(*Alloc								)(LONG size);
	void						(*Free								)(void *data);
	void						(*Print								)(_String *str);

	Bool						(*FExist							)(_Filename *name, Bool isdir);
	Bool						(*SearchFile					)(_Filename *directory, _Filename *name, _Filename *found);
	Bool						(*FKill								)(_Filename *name, Bool isdir);
	Bool						(*FCopyFile						)(_Filename *source, _Filename *dest, Bool overwrite);
	Bool						(*FRename							)(_Filename *source, _Filename *dest);
	Bool						(*FCreateDir					)(_Filename *name);
	Bool						(*ExecuteFile					)(_Filename *path);
	_Filename*			(*GetStartupPath			)(void);
	Bool						(*ExecuteProgram      )(_Filename *program, _Filename *file);

	void						(*ShowMouse						)(LONG v);
	void						(*GetSysTime					)(LONG *year, LONG *month, LONG *day, LONG *hour, LONG *minute, LONG *second);
	LONG						(*GetTimer						)(void);
	void						(*GetLineEnd					)(_String *str);
	LONG						(*GetDefaultFPS				)(void);
	LONG						(*OutString						)(_String *str, LONG flags);
	LONG						(*GetCurrentOS				)(void);
	LONG						(*GetByteOrder				)(void);
	void						(*GetGray							)(LONG *r, LONG *g, LONG *b);
	Bool						(*ChooseColor					)(Vector *col);
	void						(*GetSerialInfoEx			)(_String *s1, _String *s2, _String *s3, _String *s4, _String *s5, _String *s6);
	Bool						(*RegisterPluginEx		)(LONG api_version, LONG type, LONG id,_String *str,void *data);
	void						(*GetNewMarker				)(ULONG *l1, ULONG *l2);
	void						(*SetMousePointer			)(LONG l);
	Bool						(*ShowBitmap1					)(_Filename *fn);
	Bool						(*ShowBitmap2					)(BaseBitmap *bm);
	void						(*StopAllThreads			)(void);
	void						(*StatusClear					)(void);
	void						(*StatusSetSpin				)(void);
	void						(*StatusSetBar				)(LONG p);
	void						(*StatusSetText				)(_String *str);
	void						(*EventAdd						)(LONG type, ULONG p1, ULONG p2);
	Bool						(*DrawViews						)(LONG flags);
	void*						(*SendModelingCommand	)(LONG command, BaseDocument *doc, BaseObject *op, _BaseContainer *bc, LONG mode);
	void						(*GetGlobalTexturePath)(LONG i, _Filename *fn);
	void						(*SetGlobalTexturePath)(LONG i, _Filename *fn);
	Bool						(*GenerateTexturePath	)(_Filename *docpath, _Filename *srcname, _Filename *dstname);
	void						(*FlushTextureEx			)(_String *name);
	void						(*FlushUnusedTextures	)(void);
	void						(*GetWorldContainer		)(_BaseContainer *bc);
	void						(*ErrorStringDialog   )(LONG type, Real x, Real y, LONG is);

	void						(*lSwap								)(void *adr, LONG cnt);
	void						(*wSwap								)(void *adr, LONG cnt);
	void						(*lIntel							)(void *adr, LONG cnt);
	void						(*wIntel							)(void *adr, LONG cnt);
	void						(*lMotor							)(void *adr, LONG cnt);
	void						(*wMotor							)(void *adr, LONG cnt);

	void						(*InsertBaseDocument	)(BaseDocument *doc);
	void						(*SetActiveDocument		)(BaseDocument *doc);
	BaseDocument*		(*GetActiveDocument		)(void);
	BaseDocument*		(*GetFirstDocument		)(void);
	void						(*KillDocument				)(BaseDocument *doc);
	Bool						(*LoadFile						)(_Filename *name);
	Bool						(*LoadDocument				)(BaseDocument *doc, _Filename *name, Bool show_error);
	Bool						(*SaveDocument				)(BaseDocument *doc, _Filename *name, Bool show_error, LONG format);
	LONG						(*RenderDocument			)(BaseDocument *doc, ProgressHook *pr, void *private_data, BaseBitmap *bmp, _BaseContainer *rdata, Bool external, Bool show_errors, BaseThread *th, LONG v_left, LONG v_top, LONG v_right, LONG v_bottom);
	Vector					(*GetColor						)(LONG i);
	Bool						(*RegisterPlugin			)(LONG api_version, LONG type, LONG id,_String *str,void *data,LONG datasize);
	void						(*GetSerialInfo				)(LONG type, _String *s1, _String *s2, _String *s3, _String *s4, _String *s5, _String *s6);
	LONG						(*GetVersionType			)(void);
	void						(*FlushTexture				)(_Filename *docpath, _String *name);
	Bool						(*ReadPluginInfo			)(LONG pluginid, CHAR *buffer, LONG size);
	Bool						(*WritePluginInfo			)(LONG pluginid, CHAR *buffer, LONG size);

	void						(*EwDrawXORLine				)(EditorWindow *win, LONG x1,LONG y1,LONG x2,LONG y2);
	void						(*EwMouseDragStart		)(EditorWindow *win, LONG button,Real mx,Real my,LONG flag);
	LONG						(*EwMouseDrag					)(EditorWindow *win, Real *mx,Real *my,_BaseContainer *channels);
	LONG						(*EwMouseDragEnd			)(EditorWindow *win);
	Bool						(*EwBfGetInputState		)(EditorWindow *win, LONG askdevice,LONG askchannel,_BaseContainer *res);
	Bool						(*EwBfGetInputEvent		)(EditorWindow *win, LONG askdevice,_BaseContainer *res);

	Bool						(*GetMovieInfo				)(_Filename *name, LONG *frames, LONG *fps);
	LONG						(*ShowPopupMenu       )(CDialog *parent,LONG screenx,LONG screeny,_BaseContainer *bc,LONG flags);

	Bool						(*RegistryAdd         )(LONG sub_id, LONG main_id, void *data);
	Bool						(*RegistryRemove      )(LONG sub_id, LONG main_id);
	Registry*				(*RegistryFind				)(LONG sub_id, LONG main_id);
	Registry*				(*RegistryFindLast		)(LONG main_id);
	Registry*				(*RegistryFindFirst		)(LONG main_id);
	Registry*				(*RegistryFindPrev		)(Registry *last);
	Registry*				(*RegistryFindNext		)(Registry *last);
	Bool						(*RegistryGetAutoID		)(LONG *id);
	Bool						(*RegistryGetData			)(Registry *reg, LONG *main_id, LONG *sub_id, void **data);
};

struct C4D_Neighbor
{
	EnumerateEdges*	(*Alloc								)(LONG pcnt, Polygon *vadr, LONG vcnt, BaseSelect *bs);
	void						(*Free								)(EnumerateEdges *nb);
	void						(*GetEdgePolys				)(EnumerateEdges *nb, LONG a, LONG b,LONG *first,LONG *second);
	void						(*GetPointPolys				)(EnumerateEdges *nb, LONG pnt, LONG **dadr, LONG *dcnt);
	LONG						(*GetEdgeCount				)(EnumerateEdges *nb);
	PolyInfo*				(*GetPolyInfo					)(EnumerateEdges *nb, LONG poly);
};

struct C4D_Painter
{
	void*						(*SendPainterCommand  )(LONG command, BaseDocument *doc, PaintTexture *tex, _BaseContainer *bc);
	Bool						(*OptimalMapping			)(PMStruct *pm_adr, Polygon *vadr, LONG vcnt, Vector *padr, LONG pcnt, Bool progressbar, _BaseContainer *bc);
};

#define C4D_RADIOSITY_ID	1000969
#define C4D_CAUSTICS_ID		1000970

struct C4D_Radiosity
{
	LONG disklevel,custom_id;

	Vector					(*ComputeDiffuseIllumination)(VP_DATA *data, VolumeData *vd, Real receive_strength, Real accuracy);
	Vector					(*ComputeBlurryRay					)(VP_DATA *data, VolumeData *vd, Real dispersion, Real accuracy, LONG min, LONG max, Ray *ray);
	Bool						(*LoadGI										)(VP_DATA *data, BaseFile *file, LONG disklevel);
	Bool						(*SaveGI										)(VP_DATA *data, BaseFile *file);
	void						(*FlushGI										)(VP_DATA *data);
};

struct C4D_Caustics
{
	LONG disklevel,custom_id;

	Vector					(*ComputeSurfaceCaustic 		)(VP_DATA *data, VolumeData *vd, Real sampleradius, LONG accuracy);
	Vector					(*ComputeVolumeCaustic 			)(VP_DATA *data, const Vector &p);
	Bool						(*LoadCS										)(VP_DATA *data, BaseFile *file, LONG disklevel);
	Bool						(*SaveCS										)(VP_DATA *data, BaseFile *file);
	void						(*FlushCS										)(VP_DATA *data);
};

struct OperatingSystem
{
	LONG version;

	C4D_General				*Ge;
	C4D_Shader				*Sh;
	C4D_HyperFile			*Hf;
	C4D_BaseContainer *Bc;
	C4D_String				*St;
	C4D_Bitmap				*Bm;
	C4D_MovieSaver		*Ms;
	C4D_BaseChannel		*Ba;
	C4D_Filename			*Fn;
	C4D_File					*Fl;
	C4D_BrowseFiles		*Bf;
	C4D_Dialog				*Cd;
	C4D_UserArea			*Cu;
	C4D_Parser				*Pr;
	C4D_Resource			*Lr;
	C4D_BaseList			*Bl;
	C4D_Tag						*Tg;
	C4D_Object				*Bo;
	C4D_Document			*Bd;
	C4D_Thread				*Bt;
	C4D_Material			*Mt;
	C4D_Texture				*Tx;
	C4D_BaseSelect		*Bs;
	C4D_Key						*Bk;
	C4D_Sequence			*Bq;
	C4D_Track					*Be;
	C4D_BaseSound			*Bu;
	C4D_BaseDraw			*Br;
	C4D_BaseView			*Bv;
	C4D_Neighbor			*Nb;
	C4D_Pool					*Pl;
	C4D_BitmapFilter	*Fi;
	C4D_Painter				*Pa;
	C4D_Radiosity			*Ra;
	C4D_Caustics			*Cs;
};

extern OperatingSystem C4DOS;

#endif
