#ifndef __C4DBASEKEY_H
#define __C4DBASEKEY_H

#include "c4d_baselist.h"
#include "operatingsystem.h"
#include "ge_vector.h"

class BaseObject;
class BaseDocument;
class BaseTrack;
class BaseSequence;

class BaseKey : public BaseList2D
{
	private:
		BaseKey(void);
		~BaseKey(void);
	public:
		BaseSequence  *GetSequence(void) { return (BaseSequence*)C4DOS.Bl->GetMain(this); }
		BaseTrack			*GetTrack   (void);
		BaseObject    *GetObject  (void);
		BaseDocument  *GetDocument(void);

		BaseKey 			*GetClone(LONG flags=0);
		Bool    			CopyTo(BaseKey *dest, LONG flags=0);
		Bool 					Edit(void); 
		void					SetData(const BaseContainer &bc);
		BaseContainer GetData(void);
		void					SetTime(const BaseTime &time);
		BaseTime			GetTime(void);
		Bool					FillKey(void) { return C4DOS.Bk->FillKey(this); }

		BaseKey *next(void) { return (BaseKey*)C4DOS.Bl->Next(this); }
		BaseKey *pred(void) { return (BaseKey*)C4DOS.Bl->Pred(this); }
};

class PSRKey : public BaseKey
{
	private:
		PSRKey(void);
		~PSRKey(void);
	public:
		void SetP(const Vector &v);
		void SetVL(const Vector &v);
		void SetVR(const Vector &v);
		Vector GetP(void);
		Vector GetVL(void);
		Vector GetVR(void);
		void SoftTangents(LONG component, BaseDocument *doc);
};

class PLAKey : public BaseKey
{
	private:
		PLAKey(void);
		~PLAKey(void);
	public:
		VariableTag *GetPointTag(void);
		VariableTag *GetHermiteTag(void);
};

class TextureKey : public BaseKey
{
	private:
		TextureKey(void);
		~TextureKey(void);
	public:
		Vector GetPos(void);
		Vector GetScale(void);
		Vector GetRot(void);
		Matrix GetMl(void);
		Matrix GetMln(void);
		void SetPos(const Vector &v);
		void SetScale(const Vector &v);
		void SetRot(const Vector &v);
		void SetMl(const Matrix &m);

		BaseMaterial *GetMaterial(void);
		void SetMaterial(BaseMaterial *mat);
};

class PluginKey : public BaseKey
{
	private:
		PluginKey(void);
		~PluginKey(void);
	public:
		LONG GetPluginID(void);
		void *GetPluginData(void);
		void SetPluginData(void *mem);
};

class MarkerKey : public BaseKey
{
	private:
		MarkerKey(void);
		~MarkerKey(void);
};

BaseKey *AllocKey(LONG type, LONG subtype=0);
#define FreeKey(v) { if (v) C4DOS.Bl->Free(v); v=NULL;}

#endif
