/////////////////////////////////////////////////////////////
// CINEMA SDK : TOOL PLUGINS															 //
/////////////////////////////////////////////////////////////
// VERSION    : CINEMA 4D V6 AMBER												 //
/////////////////////////////////////////////////////////////
// (c) 1989-2001 MAXON Computer GmbH, all rights reserved	 //
/////////////////////////////////////////////////////////////

#ifndef __C4D_TOOLPLUGIN_H
#define __C4D_TOOLPLUGIN_H

#include "ge_math.h"

class _String;
class _BaseContainer;
class BaseContainer;
class BaseBitmap;
class BaseDocument;
class BaseObject;
class BaseDraw;
class BaseDrawHelp;
class BaseThread;
class CDialog;
class String;

class EditorWindow
{
	private:
		EditorWindow(void);
		~EditorWindow(void);

	public:
		void DrawXORLine(LONG x1,LONG y1,LONG x2,LONG y2);

		void MouseDragStart(LONG button,Real mx,Real my,LONG flag); // button: BFM_INPUT_xxx, flags: MOUSEDRAG_xxx
		LONG MouseDrag(Real *mx,Real *my,BaseContainer *channels); // result: MOUSEDRAG_xxx
		LONG MouseDragEnd(void);

		Bool BfGetInputState(LONG askdevice,LONG askchannel,BaseContainer *res); // device: BFM_INPUT_MOUSE,BFM_INPUT_KEYBOARD, channel: BFM_INPUT_MOUSE_xxx
		Bool BfGetInputEvent(LONG askdevice,BaseContainer *res);                 // device: BFM_INPUT_MOUSE,BFM_INPUT_KEYBOARD,

		void StatusSetText(const String &str);
};

// MouseDrag channel values
enum
{
	PENPRESSURE          = 1,
	PENTILT              = 2,
	PENROTATION					 = 3,
	PENDRAWROTATION	     = 4,
	PENDRAWRANDOMNESS    = 5,
	PENDRAWWHEELROTATION = 6,
	PENDRAWWHEELPRESSURE = 7,
	PENDRAWDISTANCE      = 8,
	PENFINGERWHEEL       = 9
};


// return values for Draw()
#define DRAW_HANDLES			  (1)
#define DRAW_AXIS					  (2)

#define DRAWFLAGS_INVERSE_Z (1)

// container id's of specific tool messages 
#define MESSAGE_TOOL_ACTIVATE   'tlac'

struct TOOLPLUGIN
{
	_String			*help;
	BaseBitmap	*icon;
	Bool				use_handles;

	Bool				(*Execute				)(BaseDocument *doc,BaseObject *op, BaseDraw *bd, EditorWindow *win, LONG button, Real mx,Real my);
	LONG				(*GetState			)(BaseDocument *doc);
	Bool				(*Message       )(const _BaseContainer *bc);  
	Bool        (*GetCursorInfo )(BaseDocument *doc, BaseDraw *bd, Real x,Real y,_BaseContainer *t_bc);
	LONG        (*Draw          )(BaseDocument *doc,BaseObject *actu_op, BaseDraw *bd, BaseDrawHelp *bh, BaseThread *bt,LONG flags);

	CDialog*		(*AllocDialog   )(_BaseContainer *bc, CDialog *cd_parent, LONG dlg_id);
	void				(*FreeDialog    )(CDialog *cd);
};

#endif
